import time
import cv2
import matplotlib.pyplot as plt
import numpy as np

from pyvcam import pvc
from pyvcam.camera import Camera
from pyvcam.constants import rgn_type

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()

    cam.meta_data_enabled = True

    cam.set_roi(0, 0, 250, 250)
    cam.set_roi(300, 100, 400, 900)
    cam.set_roi(750, 300, 200, 200)

    NUM_FRAMES = 1

    cnt = 0

    cam.start_seq(exp_time=20, num_frames=NUM_FRAMES)
    while cnt < NUM_FRAMES:
        frame, fps, frame_count = cam.poll_frame()

        num_rois = len(frame['pixel_data'])
        for roi_index in range(num_rois):
            low = np.amin(frame['pixel_data'][roi_index])
            high = np.amax(frame['pixel_data'][roi_index])
            average = np.average(frame['pixel_data'][roi_index])

            s1 = frame['meta_data']['roi_headers'][roi_index]['roi']['s1']
            s2 = frame['meta_data']['roi_headers'][roi_index]['roi']['s2']
            p1 = frame['meta_data']['roi_headers'][roi_index]['roi']['p1']
            p2 = frame['meta_data']['roi_headers'][roi_index]['roi']['p2']
            print('ROI:{:2} ({:4}, {:4}) ({:4}, {:4})\tMin:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}'.format(roi_index, s1, p1, s2, p2, low, high, average, fps, frame_count))

            alpha = 2.0  # Contrast control (1.0-3.0)
            beta = 10  # Brightness control (0-100)
            img = cv2.convertScaleAbs(frame['pixel_data'][roi_index], alpha=alpha, beta=beta)

            window_name = 'Region: {}'.format(roi_index)
            cv2.imshow(window_name, img)
            cv2.moveWindow(window_name, s1, p1)

        cnt += 1

        cv2.waitKey(0)

    cam.finish()
    cam.close()
    pvc.uninit_pvcam()

    print('Total frames: {}\n'.format(cnt))


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
from typing import Optional, Union

from .metadata import XMLMetadata
from navigate import __version__, __commit__


class OMETIFFMetadata(XMLMetadata):
    """Metadata for OME-TIFF files.

    Note
    ----
        OME-XML spec at
        https://docs.openmicroscopy.org/ome-model/6.3.1/ome-xml/index.html.
    """

    def ome_tiff_xml_dict(
        self,
        c: int = 0,
        t: int = 0,
        file_name: Union[str, list, None] = None,
        uid: Union[str, list, None] = None,
        views: Optional[list] = None,
        **kw,
    ):
        """
        Generates dictionary with same hierarchical structure as OME-XML. Useful for
        OME-TIFF and OME-XML.

        Parameters
        ----------
        c : int, optional
            Channel index, by default 0
        t : int, optional
            Time point index, by default 0
        file_name : Union[str, list, None], optional
            File name or list of file names, by default None
        uid : Union[str, list, None], optional
            Unique identifier or list of unique identifiers, by default None
        views : Optional[list], optional
            List of views, by default None
        kw : dict, optional
            Additional keyword arguments, by default None

        Returns
        -------
        ome_dict
            OME TIFF metadata dictionary
        """
        ome_dict = {
            "Creator": f"Navigate,v{__version__}, "
            f"Commit {__commit__}, "
            f"Dean Lab at UTSW",
            "xmlns": "http://www.openmicroscopy.org/Schemas/OME/2016-06",
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xsi:schemaLocation": "http://www.openmicroscopy.org/Schemas/OME/2016-06 "
            "http://www.openmicroscopy.org/Schemas/OME/2016-06/ome.xsd",
        }
        if uid is not None:
            # Assume uid is a list passed in the order of the channels
            if not isinstance(uid, list):
                uid = [uid]
            ome_dict["UUID"] = "urn:uuid:" + uid[c]
        idx = c + t * self.shape_c
        ome_dict["Image"] = {"ID": f"Image:{idx}"}
        if file_name is not None:
            # Assume file name is a list passed in the order of the channels
            if not isinstance(file_name, list):
                file_name = [file_name]
            ome_dict["Image"]["Name"] = os.path.basename(file_name[c])
        ome_dict["Image"]["Pixels"] = {"ID": f"Pixels:{idx}"}
        ome_dict["Image"]["Pixels"]["BigEndian"] = "false"
        ome_dict["Image"]["Pixels"]["Interleaved"] = "false"
        ome_dict["Image"]["Pixels"][
            "Type"
        ] = "uint16"  # Hardcoded from SharedNDArray call

        ome_dict["Image"]["Pixels"]["SizeX"] = self.shape_x
        ome_dict["Image"]["Pixels"]["SizeY"] = self.shape_y

        # The following two are commented since we split our TIFFs into one TIFF stack
        # per channel per time point
        ome_dict["Image"]["Pixels"]["SizeT"] = self.shape_t
        ome_dict["Image"]["Pixels"]["SizeC"] = self.shape_c

        ome_dict["Image"]["Pixels"]["DimensionOrder"] = "XYZCT"
        # z_steps = 1
        # if (
        #     self.configuration["experiment"]["MicroscopeState"]["image_mode"]
        #     == "z-stack"
        # ):
        #     z_steps = int(
        #         self.configuration["experiment"]["MicroscopeState"]["number_z_steps"]
        #     )
        #     ome_dict["Image"]["Pixels"]["PhysicalSizeZ"] = float(
        #         self.configuration["experiment"]["MicroscopeState"]["step_size"]
        #     )
        # elif (
        #     self.configuration["experiment"]["MicroscopeState"]["image_mode"]
        #     == "confocal-projection"
        # ):
        #     z_steps = int(
        #         self.configuration["experiment"]["MicroscopeState"]["n_plane"]
        #     )
        #     ome_dict["Image"]["Pixels"]["PhysicalSizeZ"] = (
        #         float(self.configuration["experiment"]["MicroscopeState"]["offset_end"])
        #         - float(
        #             self.configuration["experiment"]["MicroscopeState"]["offset_start"]
        #         )
        #     ) / (z_steps - 1)

        ome_dict["Image"]["Pixels"]["SizeZ"] = self.shape_z

        zoom = self.configuration["experiment"]["MicroscopeState"]["zoom"]
        pixel_size = float(
            self.configuration["configuration"]["microscopes"][self.active_microscope][
                "zoom"
            ]["pixel_size"][zoom]
        )
        (
            ome_dict["Image"]["Pixels"]["PhysicalSizeX"],
            ome_dict["Image"]["Pixels"]["PhysicalSizeY"],
        ) = (pixel_size, pixel_size)

        ome_dict["Image"]["Pixels"]["Channel"] = []
        for i in range(self.shape_c):
            d = {"ID": f"Channel:{idx}:{i}", "SamplesPerPixel": "1", "LightPath": {}}
            ome_dict["Image"]["Pixels"]["Channel"].append(d)

        if file_name is not None and uid is not None:
            ome_dict["Image"]["Pixels"]["TiffData"] = []
            if not isinstance(file_name, list):
                file_name = [file_name]
            if not isinstance(uid, list):
                uid = [uid]
            if len(file_name) == len(uid):
                # Assume file name is a list passed in the order of the channels
                for i, fn in enumerate(file_name):
                    d = {
                        "FirstC": str(i),
                        "FirstT": str(t),
                        "FirstZ": "0",
                        "IFD": "0",
                        "PlaneCount": str(self.shape_z),
                    }
                    d["UUID"] = {
                        "FileName": os.path.basename(fn),
                        "text": "urn:uuid:" + uid[i],
                    }
                    ome_dict["Image"]["Pixels"]["TiffData"].append(d)
            else:
                ome_dict["Image"]["Pixels"]["MetadataOnly"] = {}  # Required filler
        else:
            ome_dict["Image"]["Pixels"]["MetadataOnly"] = {}  # Required filler

        dt = float(
            self.configuration["experiment"]["MicroscopeState"]["timepoint_interval"]
        )
        ome_dict["Image"]["Pixels"]["TimeIncrement"] = dt

        # TODO: Populate plane positions in OME-XML
        if views is not None:
            ome_dict["Image"]["Pixels"]["Plane"] = []
            for i in range(self.shape_c):
                view_idx = i * self.shape_z
                view = views[view_idx]
                x, y, z = view["x"], view["y"], view["z"]
                # Allow additional axes (e.g. f) to couple onto existing axes (e.g. z)
                # if they are both moving along the same physical dimension
                if self._coupled_axes is not None:
                    for leader, follower in self._coupled_axes:
                        view[leader] += view[follower] / float(
                            getattr(self, f"d{follower.lower()}")
                        )

                d = {
                    "DeltaT": dt,
                    "TheT": "0",
                    "TheC": str(i),
                    "TheZ": "0",
                    "PositionX": x,
                    "PositionY": y,
                    "PositionZ": z,
                }
                ome_dict["Image"]["Pixels"]["Plane"].append(d)

        return ome_dict

    def write_xml(
        self,
        file_name: str,
        file_type: str = "OME-TIFF",
        root: Optional[str] = "OME",
        **kw,
    ) -> None:
        """Write OME-XML to file.

        Parameters
        ----------
        file_name : str
            File name
        file_type : str, optional
            File type, by default "OME-TIFF"
        root : Optional[str], optional
            Root element, by default "OME"
        kw : dict, optional
            Additional keyword arguments, by default None
        """
        return super().write_xml(file_name, file_type=file_type, root=root, **kw)

    def to_xml(
        self, file_type: str = "OME-TIFF", root: Optional[str] = "OME", **kw
    ) -> str:
        """Convert OME-XML to string.

        Parameters
        ----------
        file_type : str, optional
            File type, by default "OME-TIFF"
        root : Optional[str], optional
            Root element, by default "OME"
        kw : dict, optional
            Additional keyword arguments, by default None

        Returns
        -------
        str
            OME-XML string
        """
        return super().to_xml(file_type, root=root, **kw)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import tkinter as tk
from tkinter import ttk
from decimal import Decimal, InvalidOperation
import logging

# Third party imports

# Local imports
from navigate.view.custom_widgets.hover import Hover

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)

REGEX_DICT = {
    "float": "(^-?$)|(^-?[0-9]+\\.?[0-9]*$)",
    "float_nonnegative": "(^$)|(^[0-9]+\\.?[0-9]*$)",
    "int": "^-?[0-9]*$",
    "int_nonnegative": "^[0-9]*$",
}

"""
Base design courtesy:
Learning Path: Python GUI Programming - A Complete Reference Guide by Alan D. Moore
and B. M. Harwani

The below classes take advantage of multiple inheritance to achieve validation.

Validation Events: none, focusin, focusout, focus, key, all
- Focus is whether user is using widget
- Key is when user is typing keystrokes

Substitution Codes for validation in tkinter: %P, %s, %i, %S, %v, %V, %W, %d
- %P = proposed value from user
- %s = value currently in field
- %i = index of text being edited. String
- %S = text being inserted or deleted
- %v = widgets validate value
- %V = event that triggered validation
- %W = widgets name in TK as string
- %d = code that indicates action attempted. 0 for delete, 1 for insert, -1 for other.

String Tkinter Validation process:
- validate sets the event that triggers callback
- validatecommand takes in command that determines if data is valid
- invalidcommand runs command given if validatecommand is False

A python callable is passed to a .register function for the widget
def someFun(): do something
def otherFun(): do something
widget = ttk.Entry()
wrapped_function = widget.register(someFun)
other_function = widget.register(otherFun)
validation_function = (wrapped_function, '%P') # Can use any amount of substitution
codes above to pass in various data
invalid = (other_function, '%P', '%s')
widget.config(
            validate='all',
            validatecommand=validation_function with args,
            invalidcommand=invalid_function with args
)

"""


class ValidatedMixin:
    """Validation functions

    These functions are used to validate the input of the widget. They are called
    by the validatecommand attribute of the widget. They return True if the input
    is valid and False if it is not. They also set the error string to be displayed
    in the widget.
    """

    # error_var
    def __init__(self, *args, error_var=None, **kwargs):
        """Initialize the ValidatedMixin

        Parameters
        ----------
        *args
            Variable length argument list
        error_var : tk.StringVar
            The variable to store the error message in
            if error_var = None, creates its own variable
        **kwargs
            Keyword arguments to pass to the parent class
        """
        #: tk.StringVar: The variable to store the error message in
        self.error = error_var or tk.StringVar()

        # Calls class that is mixed in with this class
        super().__init__(*args, **kwargs)

        # History for each widgets to undo and redo
        #: list: The undo history of the widget
        self.undo_history = []
        #: list: The redo history of the widget
        self.redo_history = []

        # Validation setup
        validcmd = self.register(self._validate)
        invalidcmd = self.register(self._invalid)

        # Tkinter widget validation setup
        # Includes all validation events keystroke and focus
        self.config(
            validate="all",
            validatecommand=(
                validcmd,
                "%P",
                "%s",
                "%S",
                "%V",
                "%i",
                "%d",
            ),
            # pass in all sub codes/data
            invalidcommand=(invalidcmd, "%P", "%s", "%S", "%V", "%i", "%d"),
        )
        #: Hover: The hover bubble for the widget
        self.hover = Hover(self, text=None, type="free")
        self.hover_flag = kwargs.get("hover_flag", False)

    def _toggle_error(self, on=False):
        """Toggle the error message

        Error handler - Customize color or what happens to widget
        Changes text to red on error

        Parameters
        ----------
        on : bool
            Whether to turn the error message on or off

        """
        self.config(foreground=("red" if on else "black"))

    def _validate(self, proposed, current, char, event, index, action):
        """Validate the input of the widget

        Parameters
        ----------
        proposed : str
            The proposed value of the widget
        current : str
            The current value of the widget
        char : str
            The character being inserted or deleted
        event : str
            The event that triggered the validation
        index : str
            The index of the character being inserted or deleted
        action : str
            The action being performed. 0 for delete, 1 for insert, -1 for other

        Returns
        -------
        bool
            True if the input is valid, False if it is not
        """
        self._toggle_error(False)  # Error is off
        self.error.set("")  # No error to start
        valid = True  # Again true means no error
        if event == "focusout":  # Leaving widget
            valid = self._focusout_validate(event=event)
            if valid:
                self.add_history(event)
            else:
               if self.undo_history:
                   self.set(self.undo_history[-1])
                   self.undo_history.pop()
                   self._toggle_error(False)
                   valid = True
        elif event == "key":  # Keystroke into widget
            valid = self._key_validate(
                proposed=proposed,
                current=current,
                char=char,
                event=event,
                index=index,
                action=action,
            )
            if valid:
                self.add_history(event)
        return valid

    def _focusout_validate(self, **kwargs):
        """Validate the input of the widget when focus is lost.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to the validation function

        Returns
        -------
        bool
            True if the input is valid, False if it is not

        """
        return True

    def _key_validate(self, **kwargs):
        """Validate the input of the widget when a key is pressed.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to the validation function

        Returns
        -------
        bool
            True if the input is valid, False if it is not

        """
        return True

    # Invalid
    def _invalid(self, proposed, current, char, event, index, action):
        """Handle the invalid input of the widget

        Parameters
        ----------
        proposed : str
            The proposed value of the widget
        current : str
            The current value of the widget
        char : str
            The character being inserted or deleted
        event : str
            The event that triggered the validation
        index : str
            The index of the character being inserted or deleted
        action : str
            The action being performed. 0 for delete, 1 for insert, -1 for other

        """
        if event == "focusout":
            self._focusout_invalid(event=event)
        elif event == "key":
            self._key_invalid(
                proposed=proposed,
                current=current,
                char=char,
                event=event,
                index=index,
                action=action,
            )

    def _focusout_invalid(self, **kwargs):
        """Handle the invalid input of the widget when focus is lost.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to the invalid function

        """
        self._toggle_error(True)

    def _key_invalid(self, **kwargs):
        """Handle the invalid input of the widget when a key is pressed.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to the invalid function

        """
        pass

    # Allows a manual check on entered values to be used whenever needed
    def trigger_focusout_validation(self):
        """Trigger the focusout validation of the widget"""
        valid = self._validate("", "", "", "focusout", "", "")
        if not valid:
            self._focusout_invalid(event="focusout")
        return valid

    def add_history(self, event):
        """Add the current value to the history of the widget.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the history addition
        """
        value = self.get()
        if value != "":
            # Don't add duplicates
            if self.undo_history and self.undo_history[-1] == value:
                pass
            else:
                self.undo_history.append(value)
                self.redo_history = []
            if len(self.undo_history) > 3:
                self.undo_history.pop(0)

    def undo(self, event):
        """Undo the last change to the widget.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the undo
        """
        if self.undo_history:
            # Get the redo value
            value = self.undo_history.pop()
            if self.redo_history and self.redo_history[-1] == self.get():
                pass
            else:
                self.redo_history.append(self.get())
            if len(self.redo_history) > 3:
                self.redo_history.pop(0)

            self.set(value)
            return True
        return False

    def redo(self, event):
        """Redo the last change that was undone.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the redo

        """
        if self.redo_history:
            value = self.get()
            # add back to undo_history
            self.undo_history.append(value)
            if len(self.undo_history) > 3:
                self.undo_history.pop(0)
            value = self.redo_history.pop()
            self.set(value)
            return True
        return False


class ValidatedEntry(ValidatedMixin, ttk.Entry):
    """A validated entry widget

    Entry class that requires Entry
    Can optionally pass in a precision, min value, max value and a boolean for
    whether the entry requires a value. The min_var, max_var, and focus_update_var
    are the same as a spinbox
    """

    def __init__(
        self,
        *args,
        precision=0,
        required=False,
        min_var=None,
        max_var=None,
        focus_update_var=None,
        min="-Infinity",
        max="Infinity",
        **kwargs,
    ):
        """Initialise the entry widget

        Parameters
        ----------
        parent : tk.Widget
            The parent widget
        precision : int, optional
            The precision of the entry, by default 0
        min_var : tk.Variable, optional
            The minimum value of the entry, by default None
        max_var : tk.Variable, optional
            The maximum value of the entry, by default None
        required : bool, optional
            Whether the entry requires a value, by default False
        focus_update_var : tk.Variable, optional
            The variable to update when focus is lost, by default None
        min : str, optional
            The minimum value of the entry, by default "-Infinity"
        max : str, optional
            The maximum value of the entry, by default "Infinity"
        **kwargs
            Keyword arguments to pass to the entry
        """
        super().__init__(*args, **kwargs)
        #: Decimal: The resolution of the entry
        self.resolution = Decimal(precision)  # Number for precision given on creation
        #: int: The precision of the entry
        self.precision = (
            self.resolution.normalize().as_tuple().exponent
        )  # Precision of number as exponent
        #: tk.StringVar: The variable to store the value of the entry in
        self.variable = kwargs.get("textvariable") or tk.StringVar()
        #: str: The minimum value of the entry
        self.min = float(min)
        #: str: The maximum value of the entry
        self.max = float(max)
        #: bool: Whether the entry requires a value
        self.required = required
        self.is_fake_focusout = False

    def set(self, value):
        """Set the value of the entry to the given value.

        Parameters
        ----------
        value : str
            The value to set the entry to

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        >>> widget.set('1')
        """
        # self.delete(0, tk.END)
        self.variable.set(value)
        # self.insert(0, value)

    def set_precision(self, prec):
        """Set the precision of the entry in decimal places.

        Given a precision it will update the spinboxes ability to handle more or less
        precision for validation. This is separate from the increment value.

        Parameters
        ----------
        prec : int
            The precision to set the entry to

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        >>> widget.set_precision(1)
        """
        self.precision = prec

    def _get_precision(self):
        """Get the precision of the entry

        Returns
        -------
        int
            The precision of the entry

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        >>> widget._get_precision()
        """
        nums_after = self.resolution.find(".")

        return (-1) * len(self.resolution[nums_after + 1 :])

    def _key_validate(self, char, index, current, proposed, action, **kwargs):
        """Validate the input of the entry when a key is pressed.

        Parameters
        ----------
        char : str
            The character being inserted or deleted
        index : int
            The index of the character being inserted or deleted
        current : str
            The current value of the entry
        proposed : str
            The proposed value of the entry
        action : str
            The action being performed
        **kwargs
            Keyword arguments to pass to the invalid function

        Returns
        -------
        bool
            Whether the input is valid

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        """
        self.is_fake_focusout = True

        # Allow deletion
        if action == "0":
            if proposed == "":
                self.is_fake_focusout = False
                return True
            self._is_valid_proposed_value(proposed)
            return True

        return self._is_valid_proposed_value(proposed)
    
    def _is_valid_proposed_value(self, proposed):
        """Validate a proposed value
        
        Returns
        -------
        bool
            True if the proposed is valid, False if not
        """
        min_val = float(self.min)
        max_val = float(self.max)

        if proposed == '-':
            self._toggle_error(True)
            self.is_fake_focusout = False
            return min_val < 0
        
        if proposed == '.':
            self._toggle_error(True)
            self.is_fake_focusout = False
            return True

        # Proposed value is a Decimal, so convert and check further
        try:
            proposed = Decimal(proposed)
        except InvalidOperation:
            self.is_fake_focusout = False
            return False
        proposed_precision = proposed.as_tuple().exponent
        if any([(proposed > max_val), (proposed_precision < self.precision)]):
            return False
        
        if proposed < min_val:
            self._toggle_error(True)
            return True

        self._toggle_error(False)

        return True

    def _focusout_validate(self, event):
        """Validate the input of the entry when focus is lost.

        If entry widget is empty set the error string and return False

        Parameters
        ----------
        event : tk.Event
            The event that triggered the validation

        Returns
        -------
        bool
            Whether the input is valid

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        """
        valid = True
        value = self.get()
        max_val = self.max
        min_val = self.min

        # Check for error upon leaving widget
        if value.strip() == "" and self.required:
            self.error.set("A value is required")
            return False
        else:
            self.error.set("")

        try:
            value = Decimal(value)
        except InvalidOperation:
            self.error.set("Invalid number string: {}".format(value))
            return False

        # check if there are range limits
        if min_val == "-Infinity" or max_val == "Infinity":
            return True

        # Checking if greater than minimum
        if value < float(min_val):
            self.error.set("Value is too low (min {})".format(min_val))
            valid = False

        # Checking if less than max
        if value > float(max_val):
            self.error.set("Value is too high (max {})".format(max_val))

        # If input is valid on focusout add to history of widget
        return valid

    def _set_focus_update_var(self, event):
        """Set the focus update variable to the current value of the entry

        Parameters
        ----------
        event : tk.Event
            The event that triggered the update

        Examples
        --------
        >>> widget = ValidatedEntry(parent)
        """
        value = self.get()
        if self.focus_update_var and not self.error.get():
            self.focus_update_var.set(value)

    def _set_minimum(self, *args):
        """Set the minimum value of the entry when focus is lost.

        Parameters
        ----------
        *args
            Arguments to pass to the invalid function
        """
        current = self.get()
        try:
            new_min = self.min_var.get()
            self.min = float(new_min)
        except (tk.TclError, ValueError):
            pass
        if not current:
            self.delete(0, tk.END)
        else:
            self.variable.set(current)
        self.trigger_focusout_validation()  # Revalidate with the new minimum

    def _set_maximum(self, *args):
        """Set the maximum value of the entry when focus is lost.

        Parameters
        ----------
        *args
            Arguments to pass to the invalid function.
        """
        current = self.get()
        try:
            new_max = self.max_var.get()
            self.max = float(new_max)
        except (tk.TclError, ValueError):
            pass
        if not current:
            self.delete(0, tk.END)
        else:
            self.variable.set(current)
        self.trigger_focusout_validation()  # Revalidate with the new maximum

    def _toggle_error(self, on=False):
        """Toggle the error state of the entry.

        Parameters
        ----------
        on : bool
            Whether to turn the error state on or off


        """
        super()._toggle_error(on)
        if on and self.hover_flag:
            self.hover.seterror(self.error.get())
        else:
            self.hover.hidetip()

    def _validate(self, proposed, current, char, event, index, action):
        """Validate the input of the widget

        Parameters
        ----------
        proposed : str
            The proposed value of the widget
        current : str
            The current value of the widget
        char : str
            The character being inserted or deleted
        event : str
            The event that triggered the validation
        index : str
            The index of the character being inserted or deleted
        action : str
            The action being performed. 0 for delete, 1 for insert, -1 for other

        Returns
        -------
        bool
            True if the input is valid, False if it is not
        """
        self._toggle_error(False)  # Error is off
        self.error.set("")  # No error to start
        valid = True  # Again true means no error
        if event == "focusout":  # Leaving widget
            valid = self._focusout_validate(event=event)
            if self.is_fake_focusout:
                self.is_fake_focusout = False
                return valid
            if valid:
                self.add_history(event)
            else:
               if self.undo_history:
                   self.is_fake_focusout = True
                   self.set(self.undo_history[-1])
                   self.undo_history.pop()
                   self._toggle_error(False)
                   valid = True
        elif event == "key":  # Keystroke into widget
            valid = self._key_validate(
                proposed=proposed,
                current=current,
                char=char,
                event=event,
                index=index,
                action=action,
            )
            if valid:
                self.add_history(event)
        return valid

    def add_history(self, event):
        """Add the current value to the history of the widget.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the history addition
        """
        value = self.get()
        if value != "":
            min_val = float(self.min)
            max_val = float(self.max)
            # Don't add duplicates
            if self.undo_history and self.undo_history[-1] == value:
                pass
            elif value != "-" and value != "." and Decimal(value) >= min_val and Decimal(value) <= max_val:
                self.undo_history.append(value)
            if len(self.undo_history) > 3:
                self.undo_history.pop(0)

    def undo(self, event):
        """Undo the last change to the widget.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the undo
        """
        if self.undo_history:
            # Get the redo value
            value = self.undo_history.pop()
            if self.redo_history and self.redo_history[-1] == self.get():
                pass
            else:
                self.redo_history.append(self.get())
            if len(self.redo_history) > 3:
                self.redo_history.pop(0)
            # self.set() will trigger one focusout event
            self.is_fake_focusout = True
            self.set(value)
            return True
        return False

    def redo(self, event):
        """Redo the last change that was undone.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the redo

        """
        if self.redo_history:
            value = self.get()
            # add back to undo_history
            self.undo_history.append(value)
            if len(self.undo_history) > 3:
                self.undo_history.pop(0)
            value = self.redo_history.pop()
            # self.set() will trigger one focusout event
            self.is_fake_focusout = True
            self.set(value)
            return True
        return False


class ValidatedCombobox(ValidatedMixin, ttk.Combobox):
    """A validated combobox.

    Parameters
    ----------
    master : tk.Widget
        The parent widget
    values : list
        The values to display in the combobox
    required : bool
        Whether the combobox requires a value
    **kwargs
        Keyword arguments to pass to the combobox
    """

    def _key_validate(self, proposed, action, **kwargs):
        """Validate the input of the combobox when a key is pressed.

        Parameters
        ----------
        proposed : str
            The proposed value of the combobox
        action : str
            The action being performed
        **kwargs
            Keyword arguments to pass to the invalid function

        Returns
        -------
        bool
            Whether the input is valid
        """
        valid = True

        # Clear field with delete or backspace
        if action == "0":
            self.set("")
            return True

        # Get list of combo values
        values = self.cget("values")

        # Check for words typed in if they match then set to that value
        matching = [x for x in values if x.lower().startswith(proposed.lower())]
        if len(matching) == 0:
            valid = False
        elif len(matching) == 1:
            self.set(matching[0])
            self.icursor(tk.END)
            valid = False
        return valid

    def _focusout_validate(self, **kwargs):
        """Validate the input of the combobox when focus is lost and set the error.

        Parameters
        ----------
        **kwargs
            Keyword arguments to pass to the invalid function

        Returns
        -------
        bool
            Whether the input is valid
        """
        valid = True
        if not self.get():
            valid = False
            self.error.set("A value is required")
        return valid


class ValidatedSpinbox(ValidatedMixin, ttk.Spinbox):
    """A spinbox that validates input and can be linked to other widgets.

    Deletion always allowed, digits always allowed, if from < 0 '-' is allowed as first
    char, if increment is decimal '.' allowed, if proposed value is greater than to
    ignore key if proposed value requires more precision than increment, ignore key
    On focus out, make sure number is a valid number string and greater than from value
    If given a min_var, max_var, or focus_update_var, then the spinbox range will
    update dynamically when those valuse are changed (can be used to link to other
    widgets)
    """

    def __init__(
        self,
        *args,
        min_var=None,
        max_var=None,
        focus_update_var=None,
        required=False,
        **kwargs,
    ):
        """Initialize the spinbox

        Parameters
        ----------
        master : tk.Widget
            The parent widget
        from_ : int or float
            The minimum value of the spinbox
        to : int or float
            The maximum value of the spinbox
        increment : int or float
            The increment of the spinbox
        min_var : tk.Variable
            A variable that will be used to update the minimum value of the spinbox
        max_var : tk.Variable
            A variable that will be used to update the maximum value of the spinbox
        focus_update_var : tk.Variable
            A variable that will be used to update the value of the spinbox when it
            loses focus
        no_negative : bool
            If True, the spinbox will not allow negative values
        no_decimal : bool
            If True, the spinbox will not allow decimal values
        required : bool
            If True, the spinbox will require a value
        precision : int
            The number of decimal places allowed in the spinbox

        """
        super().__init__(*args, **kwargs)
        #: Decimal: The resolution of the spinbox
        self.resolution = str(kwargs.get("increment", "1.0"))  # Number put into spinbox
        #: int: The precision of the spinbox
        self.precision = self._get_precision()
        #: tk.StringVar: The variable to store the value of the spinbox in
        self.variable = kwargs.get("textvariable") or tk.DoubleVar()
        #: bool: Whether the spinbox requires a value
        self.required = required

        # Dynamic range checker
        if min_var:
            #: str: The minimum value of the spinbox
            self.min_var = min_var
            self.min_var.trace_add("write", self._set_minimum)
        if max_var:
            #: str: The maximum value of the spinbox
            self.max_var = max_var
            self.max_var.trace_add("write", self._set_maximum)
        self.focus_update_var = focus_update_var
        self.bind("<FocusOut>", self._set_focus_update_var)

    def set_precision(self, prec):
        """Set the precision of the spinbox in decimal places.

        Given a precision it will update the spinboxes ability to handle more or less
        precision for validation. This is separate from the increment value.

        Parameters
        ----------
        prec : int
            The number of decimal places allowed in the spinbox

        Examples
        --------
        >>> spinbox.set_precision(2)
        """
        self.precision = prec

    def _get_precision(self):
        """Get the precision of the spinbox in decimal places.

        Returns
        -------
        int
            The number of decimal places allowed in the spinbox

        Examples
        --------
        >>> spinbox._get_precision()
        """

        nums_after = self.resolution.find(".")
        return (-1) * len(self.resolution[nums_after + 1 :])

    def _key_validate(self, char, index, current, proposed, action, **kwargs):
        """Validate the key pressed in the spinbox.

        Parameters
        ----------
        char : str
            The character of the key pressed
        index : int
            The index of the cursor
        current : str
            The current value of the spinbox
        proposed : str
            The proposed value of the spinbox
        action : str
            The action of the key pressed
        **kwargs
            Additional keyword arguments

        Returns
        -------
        bool
            True if the key is valid, False if not

        Examples
        --------
        >>> spinbox._key_validate('1', 0, '0', '1', 'insert')
        """
        # Allow deletion
        if action == "0":
            if proposed == "":
                return True
            self._is_valid_proposed_value(proposed)
            return True

        return self._is_valid_proposed_value(proposed)
    
    def _is_valid_proposed_value(self, proposed):
        """Validate a proposed value
        
        Returns
        -------
        bool
            True if the proposed is valid, False if not
        """
        min_val = self.cget("from")
        max_val = self.cget("to")

        if proposed == "-" or proposed == ".":
            self._toggle_error(True)
            return True

        # Proposed value is a Decimal, so convert and check further
        try:
            proposed = Decimal(proposed)
        except InvalidOperation:
            return False
        proposed_precision = proposed.as_tuple().exponent

        if any([(proposed > max_val), (proposed_precision < self.precision)]):
            return False
        
        if proposed < min_val:
            self._toggle_error(True)
            return True

        self._toggle_error(False)

        return True

    def _focusout_validate(self, event):
        """Validate the spinbox when it loses focus and set the error message.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the validation

        Returns
        -------
        bool
            True if the spinbox is valid, False if not

        Examples
        --------
        >>> spinbox._focusout_validate()
        """
        valid = True
        value = self.get()

        max_val = self.cget("to")
        min_val = self.cget("from")
        try:
            max_val = Decimal(str(max_val))
            min_val = Decimal(str(min_val))
        except InvalidOperation:
            err_str = f"Either {min_val} or {max_val} couldn't be cast to a Decimal."
            logger.warning(err_str)
            print(err_str)

        # Check for error upon leaving widget
        if value.strip() == "" and self.required:
            self.error.set("A value is required")
            return False
        else:
            self.error.set("")

        try:
            value = Decimal(str(value))
        except InvalidOperation:
            self.error.set("Invalid Number Provided: {}".format(value))
            return False

        # Checking if greater than minimum
        if value < min_val:
            self.error.set(f"Minimum Value: {min_val:.3f}")
            valid = False

        # Checking if less than max
        if value > max_val:
            self.error.set(f"Maximum Value: {max_val:.3f}")
            valid = False

        return valid

    # Gets current value of widget and if focus_update_var is present it sets it to the
    # same value
    def _set_focus_update_var(self, event):
        """Set the focus update variable to the current value of the spinbox.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the function

        Examples
        --------
        >>> spinbox._set_focus_update_var(event)
        """
        value = self.get()
        if self.focus_update_var and not self.error.get():
            self.focus_update_var.set(value)

    def _set_minimum(self, *args):
        """Set the minimum value of the spinbox when focus is lost.

        Parameters
        ----------
        *args
            Additional arguments

        Examples
        --------
        >>> spinbox._set_minimum()
        """
        current = self.get()
        try:
            new_min = self.min_var.get()
            self.config(from_=new_min)
        except (tk.TclError, ValueError):
            pass
        if not current:
            self.delete(0, tk.END)
        else:
            self.variable.set(current)
        self.trigger_focusout_validation()  # Revalidate with the new minimum

    def _set_maximum(self, *args):
        """Set the maximum value of the spinbox when focus is lost.

        Parameters
        ----------
        *args
            Additional arguments

        Examples
        --------
        >>> spinbox._set_maximum()
        """
        current = self.get()
        try:
            new_max = self.max_var.get()
            self.config(to=new_max)
        except (tk.TclError, ValueError):
            pass
        if not current:
            self.delete(0, tk.END)
        else:
            self.variable.set(current)
        self.trigger_focusout_validation()  # Revalidate with the new maximum

    def _toggle_error(self, on=False):
        """Toggle the error message of the spinbox.

        Parameters
        ----------
        on : bool, optional
            Whether to turn the error on or off, by default False

        Examples
        --------
        >>> spinbox._toggle_error()
        """
        super()._toggle_error(on)
        if on and self.hover_flag:
            self.hover.seterror(self.error.get())
        else:
            self.hover.hidetip()

    def add_history(self, event):
        """Add the current value to the history of the widget.

        Parameters
        ----------
        event : tk.Event
            The event that triggered the history addition
        """
        value = self.get()
        if value != "":
            min_val = self.cget("from")
            max_val = self.cget("to")
            # Don't add duplicates
            if self.undo_history and self.undo_history[-1] == value:
                pass
            elif value != "-" and value != "." and Decimal(value) >= min_val and Decimal(value) <= max_val:
                self.undo_history.append(value)
                self.redo_history = []
            if len(self.undo_history) > 3:
                self.undo_history.pop(0)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import time
import ctypes

# Third Party Imports
import numpy as np
from tifffile import TiffFile, TiffFileError

# Local Imports
from navigate.model.analysis import camera
from navigate.model.devices.camera.camera_base import CameraBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SyntheticCameraController:
    """SyntheticCameraController. Synthetic Camera API."""

    def __init__(self):
        """Initialize SyntheticCameraController class."""
        pass

    def get_property_value(self, name):
        """Provides the idprop value after looking it up in the property_dict

        Parameters
        ----------
        name : str
            Not currently used.

        Returns
        -------
        property_value : int
            Currently hard-coded to -1.
        """
        return -1

    def set_property_value(self, name, value):
        """Set the value of a camera property.

        Parameters
        ----------
        name : str
            Name of the property to set.
        value : int
            Value to set the property to.
        """
        logger.debug(f"set camera property {name}: {value}")


class SyntheticCamera(CameraBase):
    """SyntheticCamera camera class."""

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize SyntheticCamera class.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocesing.managers.DictProxy
            Global configuration of the microscope
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: bool: Whether the camera is currently acquiring
        self.is_acquiring = False
        #: int: mean background count for synthetic image
        self._mean_background_count = 100
        #: float: noise sigma for synthetic image
        self._noise_sigma = camera.compute_noise_sigma()
        #: int: current image id
        self.current_frame_idx = None
        #: object: data buffer
        self.data_buffer = None
        #: int: number of frames
        self.num_of_frame = None
        #: int: previous image id
        self.pre_frame_idx = None
        #: bool: whether to use random image
        self.random_image = True
        #: int: serial number
        self.serial_number = "synthetic"
        #: float: exposure time
        self.camera_exposure_time = 0.2

        logger.info("SyntheticCamera Class Initialized")

    def __del__(self):
        """Delete SyntheticCamera class."""
        logger.info("SyntheticCamera Shutdown")
        pass

    def report_settings(self):
        """Print Camera Settings."""
        pass

    def close_camera(self):
        """Close SyntheticCamera Camera"""
        pass

    def set_sensor_mode(self, mode):
        """Set SyntheticCamera sensor mode.

        Parameters
        ----------
        mode : str
            'Normal' or 'Light-Sheet'
        """
        pass

    def set_exposure_time(self, exposure_time):
        """Set SyntheticCamera exposure time.

        All of our units are in milliseconds. Function converts to seconds.

        Parameters
        ----------
        exposure_time : float
            Exposure time in milliseconds.

        """
        self.camera_exposure_time = exposure_time / 1000

    def set_line_interval(self, line_interval_time):
        """Set SyntheticCamera line interval.

        Parameters
        ----------
        line_interval_time : float
            Line interval duration.
        """
        pass

    def set_binning(self, binning_string):
        """Set SyntheticCamera binning mode.

        Parameters
        ----------
        binning_string : str
            Desired binning properties (e.g., '2x2', '4x4', '8x8'
        """
        #: int: x binning
        self.x_binning = int(binning_string[0])
        #: int: y binning
        self.y_binning = int(binning_string[2])
        self.x_pixels = int(self.x_pixels / self.x_binning)
        self.y_pixels = int(self.y_pixels / self.y_binning)

    def initialize_image_series(self, data_buffer=None, number_of_frames=100):
        """Initialize SyntheticCamera image series.

        Parameters
        ----------
        data_buffer : int
            Size of the data to buffer.  Default is None.
        number_of_frames : int
            Number of frames.  Default is 100.
        """
        self.data_buffer = data_buffer
        self.num_of_frame = number_of_frames
        self.current_frame_idx = 0
        self.pre_frame_idx = 0
        self.is_acquiring = True

    def close_image_series(self):
        """Close image series.

        Stops the acquisition and sets is_acquiring flag to False.
        """
        self.pre_frame_idx = 0
        self.current_frame_idx = 0
        self.is_acquiring = False

    def load_images(self, filenames=None, ds=None):
        """Pre-populate the buffer with images. Can either come from TIFF files or
        Numpy stacks."""
        self.random_image = False
        #: int: current image id
        self.img_id = 0
        #: int: current tif id
        self.current_tif_id = 0
        #: list: list of tif images
        self.tif_images = []
        idx = 0
        if filenames is not None:
            # Load TIFF file into buffer as slices
            for image_file in filenames:
                try:
                    tif = TiffFile(image_file)
                    if len(tif.pages) == 1:
                        self.tif_images.append([tif.asarray()])
                    else: 
                        self.tif_images.append(tif.asarray())
                    idx += len(tif.pages)
                    if idx >= self.num_of_frame:
                        return
                except TiffFileError:
                    pass
        elif ds is not None:
            # Load a Numpy stack into buffer as slices
            # Assume the stack is in the order ZYX
            for idx, data in enumerate(ds):
                self.tif_images.append(data)
                idx += len(data)
                if idx >= self.num_of_frame:
                    return
        else:
            self.random_image = True

    def generate_new_frame(self):
        """Generate a synthetic image."""
        if not self.is_acquiring:
            return
        if self.random_image:
            image = np.random.normal(
                0,
                self._noise_sigma
                / 0.47,  # TODO: Don't hardcode 0.47 electrons per count
                size=(self.x_pixels, self.y_pixels),
            ).astype(np.uint16) + int(self._mean_background_count)
        else:
            image = self.tif_images[self.current_tif_id][self.img_id]
            self.img_id += 1
            if self.img_id >= len(self.tif_images[self.current_tif_id]):
                self.img_id = 0
                self.current_tif_id = (self.current_tif_id + 1) % len(self.tif_images)

        ctypes.memmove(
            self.data_buffer[self.current_frame_idx].ctypes.data,
            image.ctypes.data,
            self.x_pixels * self.y_pixels * 2,
        )

        self.current_frame_idx = (self.current_frame_idx + 1) % self.num_of_frame

    def get_new_frame(self):
        """Get frame from SyntheticCamera camera."""

        time.sleep(self.camera_exposure_time / 1000)
        timeout = 500
        while self.pre_frame_idx == self.current_frame_idx and timeout:
            time.sleep(0.001)
            timeout -= 1
        if timeout <= 0:
            return []
        if self.pre_frame_idx < self.current_frame_idx:
            frames = list(range(self.pre_frame_idx, self.current_frame_idx))
        else:
            frames = list(range(self.pre_frame_idx, self.num_of_frame))
            frames += list(range(0, self.current_frame_idx))
        self.pre_frame_idx = self.current_frame_idx
        logger.debug(f"Get a new frame from camera, {frames}")
        return frames

    def set_ROI(self, roi_height=2048, roi_width=2048):
        """Change the size of the active region on the camera.

        Parameters
        ----------
        roi_height : int
            Height of active camera region.
        roi_width : int
            Width of active camera region.
        """
        self.x_pixels = roi_width
        self.y_pixels = roi_height

    def get_minimum_waiting_time(self):
        """Get minimum waiting time for SyntheticCamera.

        Returns
        -------
        float
            Minimum waiting time.
        """
        return 0.01

    def calculate_readout_time(self):
        """Calculate duration of time needed to readout an image. Calculates the readout
        time and maximum frame rate according to the camera configuration settings.

        Returns
        -------
        readout_time : float
            Duration of time needed to readout an image.
        max_frame_rate : float
            Maximum framerate for a given camera acquisition mode.

        """
        exposure_time = self.camera_controller.get_property_value("exposure_time")
        readout_time = 0.01  # 10 milliseconds.
        max_frame_rate = 1 / (exposure_time + readout_time)
        return readout_time, max_frame_rate



import ctypes
from enum import IntEnum

__dll = ctypes.WinDLL(r"C:\Program Files\Mad City Labs\NanoDrive\Madlib.dll")


class MadlibErrorCode(IntEnum):
    MCL_GENERAL_ERROR = -1
    MCL_DEV_ERROR = -2
    MCL_DEV_NOT_ATTACHED = -3
    MCL_USAGE_ERROR = -4
    MCL_DEV_NOT_READY = -5
    MCL_ARGUMENT_ERROR = -6
    MCL_INVALID_AXIS = -7
    MCL_INVALID_HANDLE = -8


MadlibErrorCodeDescription = {
    MadlibErrorCode.MCL_GENERAL_ERROR: "General error",
    MadlibErrorCode.MCL_DEV_ERROR: "Device error",
    MadlibErrorCode.MCL_DEV_NOT_ATTACHED: "Device not attached",
    MadlibErrorCode.MCL_USAGE_ERROR: "Usage error",
    MadlibErrorCode.MCL_DEV_NOT_READY: "Device not ready",
    MadlibErrorCode.MCL_ARGUMENT_ERROR: "Invalid argument",
    MadlibErrorCode.MCL_INVALID_AXIS: "Invalid axis",
    MadlibErrorCode.MCL_INVALID_HANDLE: "Invalid axis handle",
}

# TODO: 'z' and 'f' have the same number? 
axes = {"x": 1, "y": 2, "z": 3, "f": 3, "aux": 4}


class MadlibError(Exception):
    """Exception for Mad City Labs device run off Madlib API."""


def in_enum(value, enum):
    values = set(item.value for item in enum)
    return value in values


def errcheck(result, func, args):
    """
    Wraps the call to DLL functions.

    Parameters
    ----------
    result : numeric or None
        Error code or positive number if successful.
    func : function
        DLL function
    args : tuple
        Arguments passed to the DLL function, defined in argtypes

    Returns
    -------
    result : int
        Error code or 0 if successful.
    """
    if result and result < 0:
        # returned a non-zero value
        if in_enum(result, MadlibErrorCode):
            raise MadlibError(MadlibErrorCodeDescription[result])
        else:
            raise Exception(f"Unknown error {result} in MCL device.")
    return result


__dll.MCL_InitHandle.restype = ctypes.c_int
__dll.MCL_InitHandle.errcheck = errcheck


def MCL_InitHandle():
    """Requests control of a single Mad City Labs Nano-Drive.

    If multiple Mad City Labs Nano-Drives are attached but not yet controlled, it is
    indeterminate which of the uncontrolled Nano-Drives this function will gain
    control of.

    Use a combination of MCL_GrabHandle, MCL_GrabAllHandles, MCL_GetAllHandles, and
    MCL_GetHandleBySerial to acquire the handle to a specific device.

    Returns
    -------
    int
        Returns a valid handle or returns 0 to indicate failure.
    """
    return int(__dll.MCL_InitHandle())


__dll.MCL_GrabAllHandles.restype = ctypes.c_int
__dll.MCL_GrabAllHandles.errcheck = errcheck


def MCL_GrabAllHandles():
    """Requests control of all of the attached  Mad City Labs Nano-Drives that are not
    yet under control.

    After calling this function use MCL_GetHandleBySerialNumber to get the handle of a
    specific device.

    Use MCL_NumberOfCurrentHandles and MCL_GetAllHandles to get a list of the handles
    acquired by this function.

    Remember that this function will take control of all of the attached Nano-Drives not
    currently under control. Some of the acquired handles may need to be released if
    those Nano-Drives are needed in other applications.

    Returns
    -------
    int
        Number of Nano-Drives currently controlled by this instance of the DLL.
    """
    return int(__dll.MCL_GrabAllHandles())


__dll.MCL_GetHandleBySerial.argtypes = [ctypes.c_short]
__dll.MCL_GetHandleBySerial.restype = ctypes.c_int
__dll.MCL_GetHandleBySerial.errcheck = errcheck


def MCL_GetHandleBySerial(serial_number):
    """Searches Nano-Drives currently controlled for a Nano-Drive whose serial number
    matches 'serial_number'.

    Parameters
    ----------
    serial_number : int
        Serial # of the Nano-Drive whose handle you want to lookup.

    Returns
    -------
    int
        Returns a valid handle or returns 0 to indicate failure
    """
    return int(__dll.MCL_GetHandleBySerial(serial_number))


__dll.MCL_ReleaseHandle.argtypes = [ctypes.c_int]
__dll.MCL_ReleaseHandle.restype = None
__dll.MCL_ReleaseHandle.errcheck = errcheck


def MCL_ReleaseHandle(handle):
    """Releases control of the specified Nano-Drive."""
    return __dll.MCL_ReleaseHandle(handle)


__dll.MCL_SingleWriteN.argtypes = [ctypes.c_double, ctypes.c_uint, ctypes.c_int]
__dll.MCL_SingleWriteN.restype = ctypes.c_int
__dll.MCL_SingleWriteN.errcheck = errcheck


def MCL_SingleWriteN(position, axis, handle):
    """Commands the Nano-Drive to move the specified axis to a position.

    Parameters
    ----------
    position : float
        Commanded position in microns.
    axis: str
        Which axis to move.  (X=1,Y=2,Z=3,AUX=4)
    handle: int
        Specifies which Nano-Drive to communicate with.

    Returns
    -------
    int
        0 if successful,  or the appropriate error code.
    """
    return int(__dll.MCL_SingleWriteN(position, axes[axis], handle))


__dll.MCL_SingleReadN.argtypes = [ctypes.c_uint, ctypes.c_int]
__dll.MCL_SingleReadN.restype = ctypes.c_double
__dll.MCL_SingleReadN.errcheck = errcheck


def MCL_SingleReadN(axis, handle):
    """Read the current position of the specified axis.

    Parameters
    ----------
    axis: str
        Which axis to read.  (X=1,Y=2,Z=3,AUX=4)
    handle: int
        Specifies which Nano-Drive to communicate with.

    Returns
    -------
    float
        Position value in microns or the appropriate error code.
    """
    return float(__dll.MCL_SingleReadN(axes[axis], handle))



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.validation import ValidatedCombobox, ValidatedSpinbox
from navigate.view.custom_widgets.hovermixin import (
    HoverButton,
    HoverCheckButton,
    HoverRadioButton,
    HoverTkButton,
)

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class LabelInput(ttk.Frame):
    """Widget class that contains label and input together.

    Acknowledgment: Based on the design from "Python GUI Programming:
    A Complete Reference Guide" by Alan D. Moore and B. M. Harwani.

    For reference on widget arguments, visit: https://tkdocs.com/shipman/

    To use kwargs for parent arguments, update kwargs like this:
    kwargs.update(width=900, height=600)
    Ensure you use super().__init__(parent, **kwargs) to override values correctly.

    User Guide:
    Create instances of LabelInput as children of a LabelFrame for grouped data.
    Access the widget within LabelInput using self.widget for customization.
    Example usage:
    LabelInput(settings, label_pos="top", label="Exposure Time",
               input_class=ttk.Spinbox, input_var=tk.IntVar(),
               input_args={"from_": 0.5, "to": 100, "increment": 0.1},
               label_args={"height": 2, "width": 4})

    """

    def __init__(
        self,
        parent,
        label_pos="left",
        label="",
        input_class=ttk.Entry,
        input_var=None,
        input_args=None,
        label_args=None,
        **kwargs,
    ):
        """Initialize LabelInput widget

        Parameters
        ----------
        parent : tk widget
            The parent widget of the LabelInput widget
        label_pos : str, optional
            The position of the label, by default "left"
        label : str, optional
            The text of the label, by default ""
        input_class : tk widget, optional
            The type of input widget, by default ttk.Entry
        input_var : tk variable, optional
            The variable of the input widget, by default None
        input_args : dict, optional
            The arguments of the input widget, by default None
        label_args : dict, optional
            The arguments of the label widget, by default None
        **kwargs : dict
            The arguments of the parent widget, by default None
        """
        # Calls frame constructor using parent and keyword args
        super().__init__(parent, **kwargs)

        # creating access point to input args for input type constructor (uses
        # these args to create the combobox etc.)
        input_args = input_args or {}

        # same for label args (args to create label etc)
        label_args = label_args or {}

        # same for variable of the input, typically this will be a Tk style var
        # like StringVar which can be accessed by any widget in the app
        #: tk.Variable: The variable of the input widget
        self.variable = input_var
        #: tk.Widget: The widget of the input widget
        self.input_class = input_class

        """ Create widgets based on their type, considering formatting differences."""
        if input_class in (
            ttk.Checkbutton,
            ttk.Button,
            ttk.Radiobutton,
            HoverButton,
            HoverTkButton,
            HoverCheckButton,
            HoverRadioButton,
        ):
            input_args["text"] = label
            input_args["variable"] = input_var
        else:
            #: ttk.Label: The label of the input widget
            self.label = ttk.Label(self, text=label, **label_args)
            self.label.grid(row=0, column=0, sticky=(tk.W + tk.E))
            input_args["textvariable"] = input_var

        """Call the passed widget type constructor with the passed args"""
        #: tk.Widget: The widget of the input widget
        self.widget = input_class(self, **input_args)

        """Specify label position"""
        if label_pos == "top":
            self.widget.grid(row=1, column=0, sticky=(tk.W + tk.E))
            self.columnconfigure(0, weight=1)
        else:
            self.widget.grid(row=0, column=1, sticky=(tk.W + tk.E))
            self.rowconfigure(0, weight=1)

    def get(self, default=None):
        """Returns the value of the input widget

        Creating a generic get function to catch all types of widgets,
        this uses the try except block for instances when tkintervariables fails

        Parameters
        ----------
        default: object, optional
            The default value to return if the get value doesn't work

        Returns
        -------
        value : str
            The value of the input widget

        Examples
        --------
        >>> value = self.get()
        """
        try:
            if self.variable:
                return self.variable.get()  # Catches widgets that don't have text
            elif isinstance(self.widget, tk.Text):
                # This is to account for the different formatting with Text widgets
                return self.widget.get("1.0", tk.END)
            else:
                return (
                    self.widget.get()
                )  # Catches all others like the normal entry widget
        except (TypeError, tk.TclError):
            # Catches times when a numeric entry input has a blank, since this
            # cannot be converted into a numeric value
            if default is not None:
                return default
            return ""

    def get_variable(self):
        """Returns the variable of the input widget

        Returns
        -------
        variable : tk variable
            The variable of the input widget

        Examples
        --------
        >>> variable = self.get_variable()
        """
        return self.variable

    def set(self, value, *args, **kwargs):
        """Sets the value of the input widget

        Parameters
        ----------
        value : str
            The value to set the input widget to

        Examples
        --------
        >>> self.set(value)
        """
        if isinstance(self.variable, tk.BooleanVar):
            # This will cast the value to bool if the variable is a Tk
            # BooleanVar, bc Boolean.Var().set() can only accept bool type
            # values
            self.variable.set(bool(value))
        elif self.variable:
            # No casting needed this will set any of the other var types
            # (String, Int, etc)
            self.variable.set(value, *args, **kwargs)
        elif type(self.widget).__name__.endswith(
            "button"
        ):  # Catches button style widgets
            # The below if statment is needed for button types because it needs
            # to be selected or enabled based on some value
            if value:
                self.widget.select()  # If there is a value select the button
            else:
                self.widget.deselect()  # If there is not deselect it
        # Catches Text type objects so the Multiline Text widget and the Entry
        # widget
        elif isinstance(self.widget, tk.Text):
            self.widget.delete("1.0", tk.END)
            # These lines are used for the specfic formatting of the Text
            # widget, 1 is the line and 0 is the char pos of that line
            self.widget.insert("1.0", value)
        elif isinstance(self.widget, ValidatedCombobox):
            self.widget.current(self.widget["values"].index(value))
        else:
            self.widget.delete(0, tk.END)
            # Basic entry widget formatting, dont need the string as the first
            # arg
            self.widget.insert(0, value)

    def set_values(self, values):
        """Set values of a combobox

        This is a function to set the values of a combobox, it will
        automatically set the values and set the first value as the current

        Values should be a list of strings or numbers that
        you want to be options in the specific widget

        Parameters
        ----------
        values : list
            list of values to be set in the widget

        Examples
        --------
        >>> widget.set_values(['a', 'b', 'c'])
        """
        if self.input_class in (
            ttk.Combobox,
            tk.Listbox,
            ttk.Spinbox,
            ValidatedCombobox,
            ValidatedSpinbox,
        ):
            self.widget["values"] = values
        else:
            print(
                "This widget class does not support list options: "
                + str(self.input_class)
            )
            logger.info(
                f"This widget class does not support list options: {self.input_class}"
            )

    def pad_input(self, left, up, right, down):
        """Pad the input widget

        This is a function to pad the input widget, it will
        automatically set the padding of the widget

        Parameters
        ----------
        left : int
            left padding
        up : int
            up padding
        right : int
            right padding
        down : int
            down padding

        Examples
        --------
        >>> widget.pad_input(10, 10, 10, 10)
        """
        self.widget.grid(padx=(left, right), pady=(up, down))



###############################################################################
# File: constants.py
# Author: Cameron Smith
# Date of Last Edit: 2020-11-10
#
# Purpose: To maintain the naming conventions used with PVCAM.h for Python
#          scripts.
#
# Notes: This file is generated by constants_generator.py. See that script for
#        details about implementation. Please do not alter this file. Instead,
#        make any additional changes to the constants_generator.py if
#        additional data is needed.
#
# Bugs: [See constants_generator.py]
###############################################################################
import ctypes,os
### DEFINES ###
MAX_CAM = 16
CAM_NAME_LEN = 32
PARAM_NAME_LEN = 32
ERROR_MSG_LEN = 255
CCD_NAME_LEN = 17
MAX_ALPHA_SER_NUM_LEN = 32
MAX_PP_NAME_LEN = 32
MAX_SYSTEM_NAME_LEN = 32
MAX_VENDOR_NAME_LEN = 32
MAX_PRODUCT_NAME_LEN = 32
MAX_CAM_PART_NUM_LEN = 32
MAX_GAIN_NAME_LEN = 32
MAX_CAM_SYSTEMS_INFO_LEN = 1024
PP_MAX_PARAMETERS_PER_FEATURE = 10
PL_MD_FRAME_SIGNATURE = 5328208
PL_MD_EXT_TAGS_MAX_SUPPORTED = 255
TYPE_INT16 = 1
TYPE_INT32 = 2
TYPE_FLT64 = 4
TYPE_UNS8 = 5
TYPE_UNS16 = 6
TYPE_UNS32 = 7
TYPE_UNS64 = 8
TYPE_ENUM = 9
TYPE_BOOLEAN = 11
TYPE_INT8 = 12
TYPE_CHAR_PTR = 13
TYPE_VOID_PTR = 14
TYPE_VOID_PTR_PTR = 15
TYPE_INT64 = 16
TYPE_SMART_STREAM_TYPE = 17
TYPE_SMART_STREAM_TYPE_PTR = 18
TYPE_FLT32 = 19
CLASS0 = 0
CLASS2 = 2
CLASS3 = 3
PARAM_DD_INFO_LENGTH = ((CLASS0<<16) + (TYPE_INT16<<24) + 1)
PARAM_DD_VERSION = ((CLASS0<<16) + (TYPE_UNS16<<24) + 2)
PARAM_DD_RETRIES = ((CLASS0<<16) + (TYPE_UNS16<<24) + 3)
PARAM_DD_TIMEOUT = ((CLASS0<<16) + (TYPE_UNS16<<24) + 4)
PARAM_DD_INFO = ((CLASS0<<16) + (TYPE_CHAR_PTR<<24) + 5)
PARAM_CAM_INTERFACE_TYPE = ((CLASS0<<16) + (TYPE_ENUM<<24) + 10)
PARAM_CAM_INTERFACE_MODE = ((CLASS0<<16) + (TYPE_ENUM<<24) + 11)
PARAM_ADC_OFFSET = ((CLASS2<<16) + (TYPE_INT16<<24)     + 195)
PARAM_CHIP_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 129)
PARAM_SYSTEM_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 130)
PARAM_VENDOR_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 131)
PARAM_PRODUCT_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 132)
PARAM_CAMERA_PART_NUMBER = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 133)
PARAM_COOLING_MODE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 214)
PARAM_PREAMP_DELAY = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 502)
PARAM_COLOR_MODE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 504)
PARAM_MPP_CAPABLE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 224)
PARAM_PREAMP_OFF_CONTROL = ((CLASS2<<16) + (TYPE_UNS32<<24)     + 507)
PARAM_PREMASK = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  53)
PARAM_PRESCAN = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  55)
PARAM_POSTMASK = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  54)
PARAM_POSTSCAN = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  56)
PARAM_PIX_PAR_DIST = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 500)
PARAM_PIX_PAR_SIZE = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  63)
PARAM_PIX_SER_DIST = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 501)
PARAM_PIX_SER_SIZE = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  62)
PARAM_SUMMING_WELL = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)   + 505)
PARAM_FWELL_CAPACITY = ((CLASS2<<16) + (TYPE_UNS32<<24)     + 506)
PARAM_PAR_SIZE = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  57)
PARAM_SER_SIZE = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  58)
PARAM_ACCUM_CAPABLE = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)   + 538)
PARAM_FLASH_DWNLD_CAPABLE = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)   + 539)
PARAM_READOUT_TIME = ((CLASS2<<16) + (TYPE_FLT64<<24)     + 179)
PARAM_CLEARING_TIME = ((CLASS2<<16) + (TYPE_INT64<<24)     + 180)
PARAM_POST_TRIGGER_DELAY = ((CLASS2<<16) + (TYPE_INT64<<24)     + 181)
PARAM_PRE_TRIGGER_DELAY = ((CLASS2<<16) + (TYPE_INT64<<24)     + 182)
PARAM_CLEAR_CYCLES = ((CLASS2<<16) + (TYPE_UNS16<<24)     +  97)
PARAM_CLEAR_MODE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 523)
PARAM_FRAME_CAPABLE = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)   + 509)
PARAM_PMODE = ((CLASS2<<16) + (TYPE_ENUM <<24)     + 524)
PARAM_TEMP = ((CLASS2<<16) + (TYPE_INT16<<24)     + 525)
PARAM_TEMP_SETPOINT = ((CLASS2<<16) + (TYPE_INT16<<24)     + 526)
PARAM_CAM_FW_VERSION = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 532)
PARAM_HEAD_SER_NUM_ALPHA = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 533)
PARAM_PCI_FW_VERSION = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 534)
PARAM_FAN_SPEED_SETPOINT = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 710)
PARAM_CAM_SYSTEMS_INFO = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 536)
PARAM_EXPOSURE_MODE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 535)
PARAM_EXPOSE_OUT_MODE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 560)
PARAM_BIT_DEPTH = ((CLASS2<<16) + (TYPE_INT16<<24)     + 511)
PARAM_IMAGE_FORMAT = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 248)
PARAM_IMAGE_COMPRESSION = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 249)
PARAM_SCAN_MODE = ((CLASS3<<16) + (TYPE_ENUM<<24)      + 250)
PARAM_SCAN_DIRECTION = ((CLASS3<<16) + (TYPE_ENUM<<24)      + 251)
PARAM_SCAN_DIRECTION_RESET = ((CLASS3<<16) + (TYPE_BOOLEAN<<24)      + 252)
PARAM_SCAN_LINE_DELAY = ((CLASS3<<16) + (TYPE_UNS16<<24)      + 253)
PARAM_SCAN_LINE_TIME = ((CLASS3<<16) + (TYPE_INT64<<24)      + 254)
PARAM_SCAN_WIDTH = ((CLASS3<<16) + (TYPE_UNS16<<24)      + 255)
PARAM_FRAME_ROTATE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 256)
PARAM_FRAME_FLIP = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 257)
PARAM_GAIN_INDEX = ((CLASS2<<16) + (TYPE_INT16<<24)     + 512)
PARAM_SPDTAB_INDEX = ((CLASS2<<16) + (TYPE_INT16<<24)     + 513)
PARAM_GAIN_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24)  + 514)
PARAM_READOUT_PORT = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 247)
PARAM_PIX_TIME = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 516)
PARAM_SHTR_CLOSE_DELAY = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 519)
PARAM_SHTR_OPEN_DELAY = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 520)
PARAM_SHTR_OPEN_MODE = ((CLASS2<<16) + (TYPE_ENUM <<24)     + 521)
PARAM_SHTR_STATUS = ((CLASS2<<16) + (TYPE_ENUM <<24)     + 522)
PARAM_IO_ADDR = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 527)
PARAM_IO_TYPE = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 528)
PARAM_IO_DIRECTION = ((CLASS2<<16) + (TYPE_ENUM<<24)      + 529)
PARAM_IO_STATE = ((CLASS2<<16) + (TYPE_FLT64<<24)     + 530)
PARAM_IO_BITDEPTH = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 531)
PARAM_GAIN_MULT_FACTOR = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 537)
PARAM_GAIN_MULT_ENABLE = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)   + 541)
PARAM_PP_FEAT_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24) +  542)
PARAM_PP_INDEX = ((CLASS2<<16) + (TYPE_INT16<<24)    +  543)
PARAM_ACTUAL_GAIN = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 544)
PARAM_PP_PARAM_INDEX = ((CLASS2<<16) + (TYPE_INT16<<24)    +  545)
PARAM_PP_PARAM_NAME = ((CLASS2<<16) + (TYPE_CHAR_PTR<<24) +  546)
PARAM_PP_PARAM = ((CLASS2<<16) + (TYPE_UNS32<<24)    +  547)
PARAM_READ_NOISE = ((CLASS2<<16) + (TYPE_UNS16<<24)     + 548)
PARAM_PP_FEAT_ID = ((CLASS2<<16) + (TYPE_UNS16<<24)    +  549)
PARAM_PP_PARAM_ID = ((CLASS2<<16) + (TYPE_UNS16<<24)    +  550)
PARAM_SMART_STREAM_MODE_ENABLED = ((CLASS2<<16) + (TYPE_BOOLEAN<<24)  +  700)
PARAM_SMART_STREAM_MODE = ((CLASS2<<16) + (TYPE_UNS16<<24)    +  701)
PARAM_SMART_STREAM_EXP_PARAMS = ((CLASS2<<16) + (TYPE_VOID_PTR<<24) +  702)
PARAM_SMART_STREAM_DLY_PARAMS = ((CLASS2<<16) + (TYPE_VOID_PTR<<24) +  703)
PARAM_EXP_TIME = ((CLASS3<<16) + (TYPE_UNS16<<24)     +   1)
PARAM_EXP_RES = ((CLASS3<<16) + (TYPE_ENUM<<24)      +   2)
PARAM_EXP_RES_INDEX = ((CLASS3<<16) + (TYPE_UNS16<<24)     +   4)
PARAM_EXPOSURE_TIME = ((CLASS3<<16) + (TYPE_UNS64<<24)     +   8)
PARAM_BOF_EOF_ENABLE = ((CLASS3<<16) + (TYPE_ENUM<<24)      +   5)
PARAM_BOF_EOF_COUNT = ((CLASS3<<16) + (TYPE_UNS32<<24)     +   6)
PARAM_BOF_EOF_CLR = ((CLASS3<<16) + (TYPE_BOOLEAN<<24)   +   7)
PARAM_CIRC_BUFFER = ((CLASS3<<16) + (TYPE_BOOLEAN<<24)   + 299)
PARAM_FRAME_BUFFER_SIZE = ((CLASS3<<16) + (TYPE_UNS64<<24)     + 300)
PARAM_BINNING_SER = ((CLASS3<<16) + (TYPE_ENUM<<24)      + 165)
PARAM_BINNING_PAR = ((CLASS3<<16) + (TYPE_ENUM<<24)      + 166)
PARAM_METADATA_ENABLED = ((CLASS3<<16) + (TYPE_BOOLEAN<<24)   + 168)
PARAM_ROI_COUNT = ((CLASS3<<16) + (TYPE_UNS16  <<24)   + 169)
PARAM_CENTROIDS_ENABLED = ((CLASS3<<16) + (TYPE_BOOLEAN<<24)   + 170)
PARAM_CENTROIDS_RADIUS = ((CLASS3<<16) + (TYPE_UNS16  <<24)   + 171)
PARAM_CENTROIDS_COUNT = ((CLASS3<<16) + (TYPE_UNS16  <<24)   + 172)
PARAM_CENTROIDS_MODE = ((CLASS3<<16) + (TYPE_ENUM   <<24)   + 173)
PARAM_CENTROIDS_BG_COUNT = ((CLASS3<<16) + (TYPE_ENUM   <<24)   + 174)
PARAM_CENTROIDS_THRESHOLD = ((CLASS3<<16) + (TYPE_UNS32  <<24)   + 175)
PARAM_TRIGTAB_SIGNAL = ((CLASS3<<16) + (TYPE_ENUM<<24)      + 180)
PARAM_LAST_MUXED_SIGNAL = ((CLASS3<<16) + (TYPE_UNS8<<24)      + 181)
PARAM_FRAME_DELIVERY_MODE = ((CLASS3<<16) + (TYPE_ENUM <<24)     + 400)

### ENUMS ###

# PL_OPEN_MODES
OPEN_EXCLUSIVE = 0

# PL_COOL_MODES
NORMAL_COOL = 0
CRYO_COOL = NORMAL_COOL + 1
NO_COOL = CRYO_COOL + 1

# PL_MPP_MODES
MPP_UNKNOWN = 0
MPP_ALWAYS_OFF = MPP_UNKNOWN + 1
MPP_ALWAYS_ON = MPP_ALWAYS_OFF + 1
MPP_SELECTABLE = MPP_ALWAYS_ON + 1

# PL_SHTR_MODES
SHTR_FAULT = 0
SHTR_OPENING = SHTR_FAULT + 1
SHTR_OPEN = SHTR_OPENING + 1
SHTR_CLOSING = SHTR_OPEN + 1
SHTR_CLOSED = SHTR_CLOSING + 1
SHTR_UNKNOWN = SHTR_CLOSED + 1

# PL_PMODES
PMODE_NORMAL = 0
PMODE_FT = PMODE_NORMAL + 1
PMODE_MPP = PMODE_FT + 1
PMODE_FT_MPP = PMODE_MPP + 1
PMODE_ALT_NORMAL = PMODE_FT_MPP + 1
PMODE_ALT_FT = PMODE_ALT_NORMAL + 1
PMODE_ALT_MPP = PMODE_ALT_FT + 1
PMODE_ALT_FT_MPP = PMODE_ALT_MPP + 1

# PL_COLOR_MODES
COLOR_NONE =  0
COLOR_RESERVED =  1
COLOR_RGGB =  2
COLOR_GRBG = COLOR_RGGB + 1
COLOR_GBRG = COLOR_GRBG + 1
COLOR_BGGR = COLOR_GBRG + 1

# PL_IMAGE_FORMATS
PL_IMAGE_FORMAT_MONO16 =  0
PL_IMAGE_FORMAT_BAYER16 = PL_IMAGE_FORMAT_MONO16 + 1
PL_IMAGE_FORMAT_MONO8 = PL_IMAGE_FORMAT_BAYER16 + 1
PL_IMAGE_FORMAT_BAYER8 = PL_IMAGE_FORMAT_MONO8 + 1
PL_IMAGE_FORMAT_MONO24 = PL_IMAGE_FORMAT_BAYER8 + 1
PL_IMAGE_FORMAT_BAYER24 = PL_IMAGE_FORMAT_MONO24 + 1
PL_IMAGE_FORMAT_RGB24 = PL_IMAGE_FORMAT_BAYER24 + 1
PL_IMAGE_FORMAT_RGB48 = PL_IMAGE_FORMAT_RGB24 + 1
PL_IMAGE_FORMAT_RGB72 = PL_IMAGE_FORMAT_RGB48 + 1
PL_IMAGE_FORMAT_MONO32 = PL_IMAGE_FORMAT_RGB72 + 1
PL_IMAGE_FORMAT_BAYER32 = PL_IMAGE_FORMAT_MONO32 + 1

# PL_IMAGE_COMPRESSIONS
PL_IMAGE_COMPRESSION_NONE =  0
PL_IMAGE_COMPRESSION_RESERVED8 =  8
PL_IMAGE_COMPRESSION_BITPACK9 = PL_IMAGE_COMPRESSION_RESERVED8 + 1
PL_IMAGE_COMPRESSION_BITPACK10 = PL_IMAGE_COMPRESSION_BITPACK9 + 1
PL_IMAGE_COMPRESSION_BITPACK11 = PL_IMAGE_COMPRESSION_BITPACK10 + 1
PL_IMAGE_COMPRESSION_BITPACK12 = PL_IMAGE_COMPRESSION_BITPACK11 + 1
PL_IMAGE_COMPRESSION_BITPACK13 = PL_IMAGE_COMPRESSION_BITPACK12 + 1
PL_IMAGE_COMPRESSION_BITPACK14 = PL_IMAGE_COMPRESSION_BITPACK13 + 1
PL_IMAGE_COMPRESSION_BITPACK15 = PL_IMAGE_COMPRESSION_BITPACK14 + 1
PL_IMAGE_COMPRESSION_RESERVED16 =  16
PL_IMAGE_COMPRESSION_BITPACK17 = PL_IMAGE_COMPRESSION_RESERVED16 + 1
PL_IMAGE_COMPRESSION_BITPACK18 = PL_IMAGE_COMPRESSION_BITPACK17 + 1
PL_IMAGE_COMPRESSION_RESERVED24 =  24
PL_IMAGE_COMPRESSION_RESERVED32 =  32

# PL_FRAME_ROTATE_MODES
PL_FRAME_ROTATE_MODE_NONE =  0
PL_FRAME_ROTATE_MODE_90CW = PL_FRAME_ROTATE_MODE_NONE + 1
PL_FRAME_ROTATE_MODE_180CW = PL_FRAME_ROTATE_MODE_90CW + 1
PL_FRAME_ROTATE_MODE_270CW = PL_FRAME_ROTATE_MODE_180CW + 1

# PL_FRAME_FLIP_MODES
PL_FRAME_FLIP_MODE_NONE =  0
PL_FRAME_FLIP_MODE_X = PL_FRAME_FLIP_MODE_NONE + 1
PL_FRAME_FLIP_MODE_Y = PL_FRAME_FLIP_MODE_X + 1
PL_FRAME_FLIP_MODE_XY = PL_FRAME_FLIP_MODE_Y + 1

# PL_PARAM_ATTRIBUTES
ATTR_CURRENT = 0
ATTR_COUNT = ATTR_CURRENT + 1
ATTR_TYPE = ATTR_COUNT + 1
ATTR_MIN = ATTR_TYPE + 1
ATTR_MAX = ATTR_MIN + 1
ATTR_DEFAULT = ATTR_MAX + 1
ATTR_INCREMENT = ATTR_DEFAULT + 1
ATTR_ACCESS = ATTR_INCREMENT + 1
ATTR_AVAIL = ATTR_ACCESS + 1

# PL_PARAM_ACCESS
ACC_READ_ONLY =  1
ACC_READ_WRITE = ACC_READ_ONLY + 1
ACC_EXIST_CHECK_ONLY = ACC_READ_WRITE + 1
ACC_WRITE_ONLY = ACC_EXIST_CHECK_ONLY + 1

# PL_IO_TYPE
IO_TYPE_TTL = 0
IO_TYPE_DAC = IO_TYPE_TTL + 1

# PL_IO_DIRECTION
IO_DIR_INPUT = 0
IO_DIR_OUTPUT = IO_DIR_INPUT + 1
IO_DIR_INPUT_OUTPUT = IO_DIR_OUTPUT + 1

# PL_READOUT_PORTS
READOUT_PORT_0 =  0
READOUT_PORT_1 = READOUT_PORT_0 + 1

# PL_CLEAR_MODES
CLEAR_NEVER = 0
CLEAR_AUTO =  CLEAR_NEVER
CLEAR_PRE_EXPOSURE = CLEAR_AUTO + 1
CLEAR_PRE_SEQUENCE = CLEAR_PRE_EXPOSURE + 1
CLEAR_POST_SEQUENCE = CLEAR_PRE_SEQUENCE + 1
CLEAR_PRE_POST_SEQUENCE = CLEAR_POST_SEQUENCE + 1
CLEAR_PRE_EXPOSURE_POST_SEQ = CLEAR_PRE_POST_SEQUENCE + 1
MAX_CLEAR_MODE = CLEAR_PRE_EXPOSURE_POST_SEQ + 1

# PL_SHTR_OPEN_MODES
OPEN_NEVER = 0
OPEN_PRE_EXPOSURE = OPEN_NEVER + 1
OPEN_PRE_SEQUENCE = OPEN_PRE_EXPOSURE + 1
OPEN_PRE_TRIGGER = OPEN_PRE_SEQUENCE + 1
OPEN_NO_CHANGE = OPEN_PRE_TRIGGER + 1

# PL_EXPOSURE_MODES
TIMED_MODE = 0
STROBED_MODE = TIMED_MODE + 1
BULB_MODE = STROBED_MODE + 1
TRIGGER_FIRST_MODE = BULB_MODE + 1
FLASH_MODE = TRIGGER_FIRST_MODE + 1
VARIABLE_TIMED_MODE = FLASH_MODE + 1
INT_STROBE_MODE = VARIABLE_TIMED_MODE + 1
MAX_EXPOSE_MODE =  7
EXT_TRIG_INTERNAL =  (7 + 0) << 8
EXT_TRIG_TRIG_FIRST =  (7 + 1) << 8
EXT_TRIG_EDGE_RISING =  (7 + 2) << 8
EXT_TRIG_LEVEL =  (7 + 3) << 8
EXT_TRIG_SOFTWARE_FIRST =  (7 + 4) << 8
EXT_TRIG_SOFTWARE_EDGE =  (7 + 5) << 8


# PL_SW_TRIG_STATUSES
PL_SW_TRIG_STATUS_TRIGGERED =  0
PL_SW_TRIG_STATUS_IGNORED = PL_SW_TRIG_STATUS_TRIGGERED + 1

# PL_EXPOSE_OUT_MODES
EXPOSE_OUT_FIRST_ROW =  0
EXPOSE_OUT_ALL_ROWS = EXPOSE_OUT_FIRST_ROW + 1
EXPOSE_OUT_ANY_ROW = EXPOSE_OUT_ALL_ROWS + 1
EXPOSE_OUT_ROLLING_SHUTTER = EXPOSE_OUT_ANY_ROW + 1
EXPOSE_OUT_LINE_TRIGGER = EXPOSE_OUT_ROLLING_SHUTTER + 1
EXPOSE_OUT_GLOBAL_SHUTTER = EXPOSE_OUT_LINE_TRIGGER + 1
MAX_EXPOSE_OUT_MODE = EXPOSE_OUT_GLOBAL_SHUTTER + 1

# PL_FAN_SPEEDS
FAN_SPEED_HIGH = 0
FAN_SPEED_MEDIUM = FAN_SPEED_HIGH + 1
FAN_SPEED_LOW = FAN_SPEED_MEDIUM + 1
FAN_SPEED_OFF = FAN_SPEED_LOW + 1

# PL_TRIGTAB_SIGNALS
PL_TRIGTAB_SIGNAL_EXPOSE_OUT = 0

# PL_FRAME_DELIVERY_MODES
PL_FRAME_DELIVERY_MODE_MAX_FPS =  0
PL_FRAME_DELIVERY_MODE_CONSTANT_INTERVALS = PL_FRAME_DELIVERY_MODE_MAX_FPS + 1

# PL_CAM_INTERFACE_TYPES
PL_CAM_IFC_TYPE_UNKNOWN =  0
PL_CAM_IFC_TYPE_1394 =  0x100
PL_CAM_IFC_TYPE_1394_A = PL_CAM_IFC_TYPE_1394 + 1
PL_CAM_IFC_TYPE_1394_B = PL_CAM_IFC_TYPE_1394_A + 1
PL_CAM_IFC_TYPE_USB =  0x200
PL_CAM_IFC_TYPE_USB_1_1 = PL_CAM_IFC_TYPE_USB + 1
PL_CAM_IFC_TYPE_USB_2_0 = PL_CAM_IFC_TYPE_USB_1_1 + 1
PL_CAM_IFC_TYPE_USB_3_0 = PL_CAM_IFC_TYPE_USB_2_0 + 1
PL_CAM_IFC_TYPE_USB_3_1 = PL_CAM_IFC_TYPE_USB_3_0 + 1
PL_CAM_IFC_TYPE_PCI =  0x400
PL_CAM_IFC_TYPE_PCI_LVDS = PL_CAM_IFC_TYPE_PCI + 1
PL_CAM_IFC_TYPE_PCIE =  0x800
PL_CAM_IFC_TYPE_PCIE_LVDS = PL_CAM_IFC_TYPE_PCIE + 1
PL_CAM_IFC_TYPE_PCIE_X1 = PL_CAM_IFC_TYPE_PCIE_LVDS + 1
PL_CAM_IFC_TYPE_PCIE_X4 = PL_CAM_IFC_TYPE_PCIE_X1 + 1
PL_CAM_IFC_TYPE_PCIE_X8 = PL_CAM_IFC_TYPE_PCIE_X4 + 1
PL_CAM_IFC_TYPE_VIRTUAL =  0x1000
PL_CAM_IFC_TYPE_ETHERNET =  0x2000


# PL_CAM_INTERFACE_MODES
PL_CAM_IFC_MODE_UNSUPPORTED =  0
PL_CAM_IFC_MODE_CONTROL_ONLY = PL_CAM_IFC_MODE_UNSUPPORTED + 1
PL_CAM_IFC_MODE_IMAGING = PL_CAM_IFC_MODE_CONTROL_ONLY + 1

# PL_CENTROIDS_MODES
PL_CENTROIDS_MODE_LOCATE =  0
PL_CENTROIDS_MODE_TRACK = PL_CENTROIDS_MODE_LOCATE + 1
PL_CENTROIDS_MODE_BLOB = PL_CENTROIDS_MODE_TRACK + 1

# PL_SCAN_MODES
PL_SCAN_MODE_AUTO =  0
PL_SCAN_MODE_PROGRAMMABLE_LINE_DELAY = PL_SCAN_MODE_AUTO + 1
PL_SCAN_MODE_PROGRAMMABLE_SCAN_WIDTH = PL_SCAN_MODE_PROGRAMMABLE_LINE_DELAY + 1

# PL_SCAN_DIRECTIONS
PL_SCAN_DIRECTION_DOWN =  0
PL_SCAN_DIRECTION_UP = PL_SCAN_DIRECTION_DOWN + 1
PL_SCAN_DIRECTION_DOWN_UP = PL_SCAN_DIRECTION_UP + 1

# PP_FEATURE_IDS
PP_FEATURE_RING_FUNCTION = 0
PP_FEATURE_BIAS = PP_FEATURE_RING_FUNCTION + 1
PP_FEATURE_BERT = PP_FEATURE_BIAS + 1
PP_FEATURE_QUANT_VIEW = PP_FEATURE_BERT + 1
PP_FEATURE_BLACK_LOCK = PP_FEATURE_QUANT_VIEW + 1
PP_FEATURE_TOP_LOCK = PP_FEATURE_BLACK_LOCK + 1
PP_FEATURE_VARI_BIT = PP_FEATURE_TOP_LOCK + 1
PP_FEATURE_RESERVED = PP_FEATURE_VARI_BIT + 1
PP_FEATURE_DESPECKLE_BRIGHT_HIGH = PP_FEATURE_RESERVED + 1
PP_FEATURE_DESPECKLE_DARK_LOW = PP_FEATURE_DESPECKLE_BRIGHT_HIGH + 1
PP_FEATURE_DEFECTIVE_PIXEL_CORRECTION = PP_FEATURE_DESPECKLE_DARK_LOW + 1
PP_FEATURE_DYNAMIC_DARK_FRAME_CORRECTION = PP_FEATURE_DEFECTIVE_PIXEL_CORRECTION + 1
PP_FEATURE_HIGH_DYNAMIC_RANGE = PP_FEATURE_DYNAMIC_DARK_FRAME_CORRECTION + 1
PP_FEATURE_DESPECKLE_BRIGHT_LOW = PP_FEATURE_HIGH_DYNAMIC_RANGE + 1
PP_FEATURE_DENOISING = PP_FEATURE_DESPECKLE_BRIGHT_LOW + 1
PP_FEATURE_DESPECKLE_DARK_HIGH = PP_FEATURE_DENOISING + 1
PP_FEATURE_ENHANCED_DYNAMIC_RANGE = PP_FEATURE_DESPECKLE_DARK_HIGH + 1
PP_FEATURE_FRAME_SUMMING = PP_FEATURE_ENHANCED_DYNAMIC_RANGE + 1
PP_FEATURE_MAX = PP_FEATURE_FRAME_SUMMING + 1

# PP_PARAMETER_IDS
PP_PARAMETER_RF_FUNCTION =  (PP_FEATURE_RING_FUNCTION * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_BIAS_ENABLED =  (PP_FEATURE_BIAS * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_BIAS_LEVEL = PP_FEATURE_BIAS_ENABLED + 1
PP_FEATURE_BERT_ENABLED =  (PP_FEATURE_BERT * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_BERT_THRESHOLD = PP_FEATURE_BERT_ENABLED + 1
PP_FEATURE_QUANT_VIEW_ENABLED =  (PP_FEATURE_QUANT_VIEW * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_QUANT_VIEW_E = PP_FEATURE_QUANT_VIEW_ENABLED + 1
PP_FEATURE_BLACK_LOCK_ENABLED =  (PP_FEATURE_BLACK_LOCK * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_BLACK_LOCK_BLACK_CLIP = PP_FEATURE_BLACK_LOCK_ENABLED + 1
PP_FEATURE_TOP_LOCK_ENABLED =  (PP_FEATURE_TOP_LOCK * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_TOP_LOCK_WHITE_CLIP = PP_FEATURE_TOP_LOCK_ENABLED + 1
PP_FEATURE_VARI_BIT_ENABLED =  (PP_FEATURE_VARI_BIT * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_VARI_BIT_BIT_DEPTH = PP_FEATURE_VARI_BIT_ENABLED + 1
PP_FEATURE_DESPECKLE_BRIGHT_HIGH_ENABLED =  (PP_FEATURE_DESPECKLE_BRIGHT_HIGH * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DESPECKLE_BRIGHT_HIGH_THRESHOLD = PP_FEATURE_DESPECKLE_BRIGHT_HIGH_ENABLED + 1
PP_FEATURE_DESPECKLE_BRIGHT_HIGH_MIN_ADU_AFFECTED = PP_FEATURE_DESPECKLE_BRIGHT_HIGH_THRESHOLD + 1
PP_FEATURE_DESPECKLE_DARK_LOW_ENABLED =  (PP_FEATURE_DESPECKLE_DARK_LOW * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DESPECKLE_DARK_LOW_THRESHOLD = PP_FEATURE_DESPECKLE_DARK_LOW_ENABLED + 1
PP_FEATURE_DESPECKLE_DARK_LOW_MAX_ADU_AFFECTED = PP_FEATURE_DESPECKLE_DARK_LOW_THRESHOLD + 1
PP_FEATURE_DEFECTIVE_PIXEL_CORRECTION_ENABLED =  (PP_FEATURE_DEFECTIVE_PIXEL_CORRECTION * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DYNAMIC_DARK_FRAME_CORRECTION_ENABLED =  (PP_FEATURE_DYNAMIC_DARK_FRAME_CORRECTION * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_HIGH_DYNAMIC_RANGE_ENABLED =  (PP_FEATURE_HIGH_DYNAMIC_RANGE * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DESPECKLE_BRIGHT_LOW_ENABLED =  (PP_FEATURE_DESPECKLE_BRIGHT_LOW * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DESPECKLE_BRIGHT_LOW_THRESHOLD = PP_FEATURE_DESPECKLE_BRIGHT_LOW_ENABLED + 1
PP_FEATURE_DESPECKLE_BRIGHT_LOW_MAX_ADU_AFFECTED = PP_FEATURE_DESPECKLE_BRIGHT_LOW_THRESHOLD + 1
PP_FEATURE_DENOISING_ENABLED =  (PP_FEATURE_DENOISING * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DENOISING_NO_OF_ITERATIONS = PP_FEATURE_DENOISING_ENABLED + 1
PP_FEATURE_DENOISING_GAIN = PP_FEATURE_DENOISING_NO_OF_ITERATIONS + 1
PP_FEATURE_DENOISING_OFFSET = PP_FEATURE_DENOISING_GAIN + 1
PP_FEATURE_DENOISING_LAMBDA = PP_FEATURE_DENOISING_OFFSET + 1
PP_FEATURE_DESPECKLE_DARK_HIGH_ENABLED =  (PP_FEATURE_DESPECKLE_DARK_HIGH * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_DESPECKLE_DARK_HIGH_THRESHOLD = PP_FEATURE_DESPECKLE_DARK_HIGH_ENABLED + 1
PP_FEATURE_DESPECKLE_DARK_HIGH_MIN_ADU_AFFECTED = PP_FEATURE_DESPECKLE_DARK_HIGH_THRESHOLD + 1
PP_FEATURE_ENHANCED_DYNAMIC_RANGE_ENABLED =  (PP_FEATURE_ENHANCED_DYNAMIC_RANGE * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_FRAME_SUMMING_ENABLED =  (PP_FEATURE_FRAME_SUMMING * PP_MAX_PARAMETERS_PER_FEATURE)
PP_FEATURE_FRAME_SUMMING_COUNT = PP_FEATURE_FRAME_SUMMING_ENABLED + 1
PP_FEATURE_FRAME_SUMMING_32_BIT_MODE = PP_FEATURE_FRAME_SUMMING_COUNT + 1
PP_PARAMETER_ID_MAX = PP_FEATURE_FRAME_SUMMING_32_BIT_MODE + 1

# PL_SMT_MODES
SMTMODE_ARBITRARY_ALL =  0
SMTMODE_MAX = SMTMODE_ARBITRARY_ALL + 1

# PL_IMAGE_STATUSES
READOUT_NOT_ACTIVE = 0
EXPOSURE_IN_PROGRESS = READOUT_NOT_ACTIVE + 1
READOUT_IN_PROGRESS = EXPOSURE_IN_PROGRESS + 1
READOUT_COMPLETE = READOUT_IN_PROGRESS + 1
FRAME_AVAILABLE =  READOUT_COMPLETE
READOUT_FAILED = FRAME_AVAILABLE + 1
ACQUISITION_IN_PROGRESS = READOUT_FAILED + 1
MAX_CAMERA_STATUS = ACQUISITION_IN_PROGRESS + 1

# PL_CCS_ABORT_MODES
CCS_NO_CHANGE =  0
CCS_HALT = CCS_NO_CHANGE + 1
CCS_HALT_CLOSE_SHTR = CCS_HALT + 1
CCS_CLEAR = CCS_HALT_CLOSE_SHTR + 1
CCS_CLEAR_CLOSE_SHTR = CCS_CLEAR + 1
CCS_OPEN_SHTR = CCS_CLEAR_CLOSE_SHTR + 1
CCS_CLEAR_OPEN_SHTR = CCS_OPEN_SHTR + 1

# PL_IRQ_MODES
NO_FRAME_IRQS =  0
BEGIN_FRAME_IRQS = NO_FRAME_IRQS + 1
END_FRAME_IRQS = BEGIN_FRAME_IRQS + 1
BEGIN_END_FRAME_IRQS = END_FRAME_IRQS + 1

# PL_CIRC_MODES
CIRC_NONE =  0
CIRC_OVERWRITE = CIRC_NONE + 1
CIRC_NO_OVERWRITE = CIRC_OVERWRITE + 1

# PL_EXP_RES_MODES
EXP_RES_ONE_MILLISEC =  0
EXP_RES_ONE_MICROSEC = EXP_RES_ONE_MILLISEC + 1
EXP_RES_ONE_SEC = EXP_RES_ONE_MICROSEC + 1

# PL_SRC_MODES
SCR_PRE_OPEN_SHTR =  0
SCR_POST_OPEN_SHTR = SCR_PRE_OPEN_SHTR + 1
SCR_PRE_FLASH = SCR_POST_OPEN_SHTR + 1
SCR_POST_FLASH = SCR_PRE_FLASH + 1
SCR_PRE_INTEGRATE = SCR_POST_FLASH + 1
SCR_POST_INTEGRATE = SCR_PRE_INTEGRATE + 1
SCR_PRE_READOUT = SCR_POST_INTEGRATE + 1
SCR_POST_READOUT = SCR_PRE_READOUT + 1
SCR_PRE_CLOSE_SHTR = SCR_POST_READOUT + 1
SCR_POST_CLOSE_SHTR = SCR_PRE_CLOSE_SHTR + 1

# PL_CALLBACK_EVENT
PL_CALLBACK_BOF =  0
PL_CALLBACK_EOF = PL_CALLBACK_BOF + 1
PL_CALLBACK_CHECK_CAMS = PL_CALLBACK_EOF + 1
PL_CALLBACK_CAM_REMOVED = PL_CALLBACK_CHECK_CAMS + 1
PL_CALLBACK_CAM_RESUMED = PL_CALLBACK_CAM_REMOVED + 1
PL_CALLBACK_MAX = PL_CALLBACK_CAM_RESUMED + 1

# PL_MD_FRAME_FLAGS
PL_MD_FRAME_FLAG_ROI_TS_SUPPORTED =  0x01
PL_MD_FRAME_FLAG_UNUSED_2 =  0x02
PL_MD_FRAME_FLAG_UNUSED_3 =  0x04
PL_MD_FRAME_FLAG_UNUSED_4 =  0x10
PL_MD_FRAME_FLAG_UNUSED_5 =  0x20
PL_MD_FRAME_FLAG_UNUSED_6 =  0x40
PL_MD_FRAME_FLAG_UNUSED_7 =  0x80


# PL_MD_ROI_FLAGS
PL_MD_ROI_FLAG_INVALID =  0x01
PL_MD_ROI_FLAG_HEADER_ONLY =  0x02
PL_MD_ROI_FLAG_UNUSED_3 =  0x04
PL_MD_ROI_FLAG_UNUSED_4 =  0x10
PL_MD_ROI_FLAG_UNUSED_5 =  0x20
PL_MD_ROI_FLAG_UNUSED_6 =  0x40
PL_MD_ROI_FLAG_UNUSED_7 =  0x80


# PL_MD_EXT_TAGS
PL_MD_EXT_TAG_PARTICLE_ID =  0
PL_MD_EXT_TAG_PARTICLE_M0 = PL_MD_EXT_TAG_PARTICLE_ID + 1
PL_MD_EXT_TAG_PARTICLE_M2 = PL_MD_EXT_TAG_PARTICLE_M0 + 1
PL_MD_EXT_TAG_MAX = PL_MD_EXT_TAG_PARTICLE_M2 + 1

### STRUCTS ###
class PVCAM_FRAME_INFO_GUID(ctypes.Structure):
    _fields_ = [
                ('f1', ctypes.c_uint32),
                ('f2', ctypes.c_uint16),
                ('f3', ctypes.c_uint16),
                ('f4', ctypes.c_uint8 * 8),
               ]

class FRAME_INFO(ctypes.Structure):
    _fields_ = [
                ('FrameInfoGUID', ctypes.c_void_p),
                ('hCam', ctypes.c_int16),
                ('FrameNr', ctypes.c_int32),
                ('TimeStamp', ctypes.c_long),
                ('ReadoutTime', ctypes.c_int32),
                ('TimeStampBOF', ctypes.c_long),
               ]

class smart_stream_type(ctypes.Structure):
    _fields_ = [
                ('entries', ctypes.c_uint16),
                ('params', ctypes.c_uint32),
               ]

class rgn_type(ctypes.Structure):
    _fields_ = [
                ('s1', ctypes.c_uint16),
                ('s2', ctypes.c_uint16),
                ('sbin', ctypes.c_uint16),
                ('p1', ctypes.c_uint16),
                ('p2', ctypes.c_uint16),
                ('pbin', ctypes.c_uint16),
               ]

class io_entry(ctypes.Structure):
    _fields_ = [
                ('io_port', ctypes.c_uint16),
                ('io_type', ctypes.c_uint32),
                ('state', ctypes.c_float),
               ]

class io_list(ctypes.Structure):
    _fields_ = [
                ('pre_open', ctypes.c_void_p),
                ('post_open', ctypes.c_void_p),
                ('pre_flash', ctypes.c_void_p),
                ('post_flash', ctypes.c_void_p),
                ('pre_integrate', ctypes.c_void_p),
                ('post_integrate', ctypes.c_void_p),
                ('pre_readout', ctypes.c_void_p),
                ('post_readout', ctypes.c_void_p),
                ('pre_close', ctypes.c_void_p),
                ('post_close', ctypes.c_void_p),
               ]

class active_camera_type(ctypes.Structure):
    _fields_ = [
                ('shutter_close_delay', ctypes.c_uint16),
                ('shutter_open_delay', ctypes.c_uint16),
                ('rows', ctypes.c_uint16),
                ('cols', ctypes.c_uint16),
                ('prescan', ctypes.c_uint16),
                ('postscan', ctypes.c_uint16),
                ('premask', ctypes.c_uint16),
                ('postmask', ctypes.c_uint16),
                ('preflash', ctypes.c_uint16),
                ('clear_count', ctypes.c_uint16),
                ('preamp_delay', ctypes.c_uint16),
                ('mpp_selectable', ctypes.c_short),
                ('frame_selectable', ctypes.c_short),
                ('do_clear', ctypes.c_int16),
                ('open_shutter', ctypes.c_int16),
                ('mpp_mode', ctypes.c_short),
                ('frame_transfer', ctypes.c_short),
                ('alt_mode', ctypes.c_short),
                ('exp_res', ctypes.c_uint32),
                ('io_hdr', ctypes.c_void_p),
               ]

class md_frame_header(ctypes.Structure):
    _fields_ = [
                ('signature', ctypes.c_uint32),
                ('version', ctypes.c_uint8),
                ('frameNr', ctypes.c_uint32),
                ('roiCount', ctypes.c_uint16),
                ('timestampBOF', ctypes.c_uint32),
                ('timestampEOF', ctypes.c_uint32),
                ('timestampResNs', ctypes.c_uint32),
                ('exposureTime', ctypes.c_uint32),
                ('exposureTimeResNs', ctypes.c_uint32),
                ('roiTimestampResNs', ctypes.c_uint32),
                ('bitDepth', ctypes.c_uint8),
                ('colorMask', ctypes.c_uint8),
                ('flags', ctypes.c_uint8),
                ('extendedMdSize', ctypes.c_uint16),
                ('imageFormat', ctypes.c_uint8),
                ('imageCompression', ctypes.c_uint8),
                ('_reserved', ctypes.c_uint8 * 6),
               ]

class md_frame_header_v3(ctypes.Structure):
    _fields_ = [
                ('signature', ctypes.c_uint32),
                ('version', ctypes.c_uint8),
                ('frameNr', ctypes.c_uint32),
                ('roiCount', ctypes.c_uint16),
                ('timestampBOF', ctypes.c_void_p),
                ('timestampEOF', ctypes.c_void_p),
                ('exposureTime', ctypes.c_void_p),
                ('bitDepth', ctypes.c_uint8),
                ('colorMask', ctypes.c_uint8),
                ('flags', ctypes.c_uint8),
                ('extendedMdSize', ctypes.c_uint16),
                ('imageFormat', ctypes.c_uint8),
                ('imageCompression', ctypes.c_uint8),
                ('_reserved', ctypes.c_uint8 * 6),
               ]

class md_frame_roi_header(ctypes.Structure):
    _fields_ = [
                ('roiNr', ctypes.c_uint16),
                ('timestampBOR', ctypes.c_uint32),
                ('timestampEOR', ctypes.c_uint32),
                ('roi', ctypes.c_void_p),
                ('flags', ctypes.c_uint8),
                ('extendedMdSize', ctypes.c_uint16),
                ('roiDataSize', ctypes.c_uint32),
                ('_reserved', ctypes.c_uint8 * 3),
               ]

class md_ext_item_info(ctypes.Structure):
    _fields_ = [
                ('tag', ctypes.c_void_p),
                ('type', ctypes.c_uint16),
                ('size', ctypes.c_uint16),
                ('name', ctypes.c_char_p),
               ]

class md_ext_item(ctypes.Structure):
    _fields_ = [
                ('tagInfo', ctypes.c_void_p),
                ('value', ctypes.c_void_p),
               ]

class md_ext_item_collection(ctypes.Structure):
    _fields_ = [
                ('count', ctypes.c_uint16),
               ]

class md_frame_roi(ctypes.Structure):
    _fields_ = [
                ('header', ctypes.c_void_p),
                ('data', ctypes.c_void_p),
                ('dataSize', ctypes.c_uint32),
                ('extMdData', ctypes.c_void_p),
                ('extMdDataSize', ctypes.c_uint16),
               ]

class md_frame(ctypes.Structure):
    _fields_ = [
                ('header', ctypes.c_void_p),
                ('extMdData', ctypes.c_void_p),
                ('extMdDataSize', ctypes.c_uint16),
                ('impliedRoi', ctypes.c_void_p),
                ('roiArray', ctypes.c_void_p),
                ('roiCapacity', ctypes.c_uint16),
                ('roiCount', ctypes.c_uint16),
               ]




# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Library Imports
import logging
import time

# Third Party Imports
import numpy as np
import serial

# Local Imports
from navigate.model.devices.filter_wheel.filter_wheel_base import FilterWheelBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def build_filter_wheel_connection(comport, baudrate, timeout=0.25):
    """Build SutterFilterWheel Serial Port connection

    Attributes
    ----------
    comport : str
        Comport for communicating with the filter wheel, e.g., COM1.
    baudrate : int
        Baud rate for communicating with the filter wheel, e.g., 9600.
    timeout : float
        Timeout for communicating with the filter wheel, e.g., 0.25.

    Returns
    -------
    serial.Serial
        Serial port connection to the filter wheel.

    Raises
    ------
    UserWarning
        Could not communicate with Sutter Lambda 10-B via COMPORT.
    """
    logging.debug(f"SutterFilterWheel - Opening Serial Port {comport}")
    try:
        return serial.Serial(comport, baudrate, timeout=timeout)
    except serial.SerialException:
        logger.warning("SutterFilterWheel - Could not establish Serial Port Connection")
        raise UserWarning(
            "Could not communicate with Sutter Lambda 10-B via COMPORT", comport
        )


class SutterFilterWheel(FilterWheelBase):
    """SutterFilterWheel - Class for controlling Sutter Lambda Filter Wheels

    Note
    ----
        Additional information on the Sutter Lambda 10-B can be found at:
        https://www.sutter.com/manuals/LB10-3_OpMan.pdf
    """

    def __init__(self, microscope_name, device_connection, configuration):
        """Initialize the SutterFilterWheel class.

        Parameters
        ----------
        microscope_name : str
            Name of the microscope.
        device_connection : dict
            Dictionary of device connections.
        configuration : dict
            Dictionary of configuration parameters.
        """
        super().__init__(microscope_name, device_connection, configuration)

        #: obj: Serial port connection to the filter wheel.
        self.serial = device_connection

        #: str: Name of the microscope.
        self.microscope_name = microscope_name

        #: int: Number of filter wheels.
        self.number_of_filter_wheels = configuration["configuration"]["microscopes"][
            microscope_name
        ]["filter_wheel"]["hardware"]["wheel_number"]

        #: bool: Wait until filter wheel has completed movement.
        self.wait_until_done = True

        #: float: Delay in s for the wait until done function.
        self.wait_until_done_delay = None

        #: bool: Read on initialization.
        self.read_on_init = True

        #: int: Filter wheel speed.
        self.speed = 2

        # Delay in s for the wait until done function
        #: np.matrix: Delay matrix for filter wheel.
        self.delay_matrix = np.matrix(
            [
                [0, 0.031, 0.051, 0.074, 0.095, 0.115],
                [0, 0.040, 0.065, 0.095, 0.120, 0.148],
                [0, 0.044, 0.075, 0.105, 0.136, 0.168],
                [0, 0.050, 0.088, 0.127, 0.165, 0.205],
                [0, 0.060, 0.108, 0.156, 0.205, 0.250],
                [0, 0.068, 0.123, 0.178, 0.235, 0.290],
                [0, 0.124, 0.235, 0.350, 0.460, 0.580],
                [0, 0.230, 0.440, 0.650, 0.860, 1.100],
            ]
        )

        logger.debug("SutterFilterWheel - Placing device In Online Mode")

        self.serial.write(bytes.fromhex("ee"))

        if self.read_on_init:
            self.read(2)  # class 'bytes'
            #: bool: Software initialization complete flag.
            self.init_finished = True
            logger.debug("SutterFilterWheel - Initialized.")
        else:
            self.init_finished = False

        # Set filter to the 0th position by default upon initialization.
        self.set_filter(list(self.filter_dictionary.keys())[0])

        logger.debug("SutterFilterWheel -  Placed in Default Filter Position.")

    def __enter__(self):
        """Enter the SutterFilterWheel context manager."""
        return self

    def __exit__(self, *args, **kwargs):
        """Exit the SutterFilterWheel context manager."""
        logger.debug("SutterFilterWheel - Closing Device.")
        self.close()

    def filter_change_delay(self, filter_name):
        """Calculate duration of time necessary to change filter wheel positions.

        Calculate duration of time necessary to change filter wheel positions.
        Identifies the number of positions that must be switched, and then retrieves
        the duration of time necessary to perform the switch from self.delay_matrix.

        Parameters
        ----------
        filter_name : str
            Name of filter that we want to move to.

        Note
        ----
            Detailed information on timing located on page 38:
            https://www.sutter.com/manuals/LB10-3_OpMan.pdf

        Warn
        ----
            Delay matrix should be model specific.

        Returns
        -------
        Index Error
            If the filter wheel position is out of range.
        """
        # Find the old and new positions, the distance between them, and the
        # delay necessary.
        old_position = self.wheel_position
        self.wheel_position = self.filter_dictionary[filter_name]
        delta_position = int(abs(old_position - self.wheel_position))
        try:
            self.wait_until_done_delay = self.delay_matrix[delta_position, self.speed]
        except IndexError:
            # Murdered by the hard coded delay matrix - Guess a value
            self.wait_until_done_delay = 0.01

    def set_filter(self, filter_name, wait_until_done=True):
        """Change the filter wheel to the filter designated by the filter
        position argument.

        Parameters
        ----------
        filter_name : str
            Name of filter to move to.
        wait_until_done : bool
            Waits duration of time necessary for filter wheel to change positions.
        """
        if self.check_if_filter_in_filter_dictionary(filter_name) is True:

            # Calculate the Delay Needed to Change the Positions
            self.filter_change_delay(filter_name)

            # Make sure you are moving it to a reasonable filter position at a
            # reasonable speed.
            assert self.wheel_position in range(10)
            assert self.speed in range(8)

            # If previously we did not confirm that the initialization was complete,
            # check now.
            if not self.init_finished:
                self.read(2)
                self.init_finished = True
                logger.debug("SutterFilterWheel - Initialized.")

            for wheel_idx in range(self.number_of_filter_wheels):
                """Loop through each filter, and send the binary sequence via serial to
                move to the desired filter wheel position
                When number_of_filter_wheels = 1, loop executes once, and only wheel A
                changes. When number_of_filter_wheels = 2, loop executes twice, with
                both wheel A and B moving to the same position sequentially
                Filter Wheel Command Byte Encoding = wheel + (self.speed*16) +
                position = command byte
                """
                logger.debug(f"SutterFilterWheel - Moving to Position {wheel_idx}")
                output_command = wheel_idx * 128 + self.wheel_position + 16 * self.speed
                output_command = output_command.to_bytes(1, "little")
                self.serial.write(output_command)

            #  Wheel Position Change Delay
            if wait_until_done:
                time.sleep(self.wait_until_done_delay)

    def read(self, num_bytes):
        """Reads the specified number of bytes from the serial port.

        Parameters
        ----------
        num_bytes : int
            Number of bytes to read from the serial port.

        Returns
        -------
        bytes
            Bytes read from the serial port.

        Raises
        ------
        UserWarning
            The serial port to the Sutter Lambda 10-B is on, but it isn't responding as
            expected.
        """
        for i in range(100):
            num_waiting = self.serial.inWaiting()
            if num_waiting == num_bytes:
                break
            time.sleep(0.02)
        else:
            logger.error(
                "The serial port to the Sutter Lambda 10-B is on, but it isn't "
                "responding as expected."
            )
            raise UserWarning(
                "The serial port to the Sutter Lambda 10-B is on, but it isn't "
                "responding as expected."
            )
        return self.serial.read(num_bytes)

    def close(self):
        """Close the SutterFilterWheel serial port.

        Sets the filter wheel to the Empty-Alignment position and then closes the port.
        """
        logger.debug("SutterFilterWheel - Closing the Filter Wheel Serial Port")
        self.set_filter(list(self.filter_dictionary.keys())[0])
        self.serial.close()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os
import logging
from typing import Optional

# Local Imports
from navigate.tools import xml_tools
from navigate import __version__, __commit__

from multiprocessing.managers import DictProxy

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class Metadata:
    def __init__(self) -> None:
        """Metadata class

        Store and convert internal representation of metadata (configuration,
        experiment, etc.) to alternative file types.

        Note
        ----
            Concept and some of the code borrowed from python-microscopy
            (https://github.com/python-microscopy/python-microscopy/).
        """
        #: dict: Configuration dictionary
        self._configuration = None

        #: float: Pixel size in x (um)
        #: float: Pixel size in y (um)
        #: float: Pixel size in z (um)
        self.dx, self.dy, self.dz = 1, 1, 1  # pixel sizes (um)

        #: float: Time displacement (s)
        self.dt = 1

        #: int: Step size between channels, should always be 1
        self.dc = 1

        self._order = "XYCZT"
        self._per_stack = True
        self._multiposition = False
        self._coupled_axes = None

        #: int: Shape of the data in x
        #: int: Shape of the data in y
        #: int: Shape of the data in z
        #: int: Shape of the data in t
        #: int: Shape of the data in c
        self.shape_x, self.shape_y, self.shape_z, self.shape_t, self.shape_c = (
            1,
            1,
            1,
            1,
            1,
        )

        #: int: Number of positions
        self.positions = 1

        #: str: Active microscope
        self.active_microscope = None

    @property
    def configuration(self) -> Optional[DictProxy]:
        """Return configuration dictionary

        Returns
        -------
        Optional[DictProxy]
            Configuration dictionary
        """
        return self._configuration

    @configuration.setter
    def configuration(self, configuration: DictProxy) -> None:
        """Set configuration dictionary

        Parameters
        ----------
        configuration : DictProxy
            Configuration dictionary
        """
        self._configuration = configuration
        self.set_from_configuration_experiment()

    @property
    def per_stack(self) -> bool:
        """Return per stack

        Returns
        -------
        bool
            True if per stack, False otherwise
        """
        return self._per_stack

    def set_from_dict(self, metadata_config: dict) -> None:
        """Set from a dictionary
        
        Parameters
        ----------
        metadata_config : dict
            dictionary of metadata: "c", "z", "t", "p", "is_dynamic", "per_stack"
        """
        self.shape_c = metadata_config.get("c", self.shape_c)
        self.shape_z = metadata_config.get("z", self.shape_z)
        self.shape_t = metadata_config.get("t", self.shape_t)
        self.positions = metadata_config.get("p", self.positions)
        if metadata_config.get("is_dynamic", False):
            self._multiposition = True
        self._per_stack = metadata_config.get("per_stack", self._per_stack)
    
    def set_from_configuration_experiment(self) -> None:
        """Set from configuration experiment"""
        if (
            self.configuration.get("experiment") is not None
            and self.configuration.get("configuration") is not None
        ):
            self.active_microscope = self.configuration["experiment"][
                "MicroscopeState"
            ]["microscope_name"]
            self.set_shape_from_configuration_experiment()
            self.set_stack_order_from_configuration_experiment()

    def set_shape_from_configuration_experiment(self) -> None:
        """Set shape from configuration experiment"""
        state = self.configuration["experiment"]["MicroscopeState"]
        scope = self.configuration["configuration"]["microscopes"][
            self.active_microscope
        ]
        zoom = state["zoom"]
        pixel_size = float(scope["zoom"]["pixel_size"][zoom])
        self.dx, self.dy = pixel_size, pixel_size
        self.dz = float(abs(state["step_size"]))
        self.dt = float(state["timepoint_interval"])

        self.shape_x = int(
            self.configuration["experiment"]["CameraParameters"]["x_pixels"]
        )
        self.shape_y = int(
            self.configuration["experiment"]["CameraParameters"]["y_pixels"]
        )
        if (
            (state["image_mode"] == "z-stack")
            or (state["image_mode"] == "customized")
            or (state["image_mode"] == "ConstantVelocityAcquisition")
        ):
            self.shape_z = int(state["number_z_steps"])
        else:
            self.shape_z = 1
        self.shape_t = int(state["timepoints"])
        self.shape_c = sum(
            [v["is_selected"] is True for k, v in state["channels"].items()]
        )

        self._multiposition = state["is_multiposition"]

        if bool(self._multiposition):
            self.positions = len(self.configuration["experiment"]["MultiPositions"])
        else:
            self.positions = 1

        # tiff
        if state["image_mode"] == "customized":
            self._multiposition = True
        # let the data sources have the ability to save more frames
        # self._multiposition = True
        # self.positions = len(
        #     self.configuration["experiment"]["MultiPositions"]
        # )

        # Allow additional axes (e.g. f) to couple onto existing axes (e.g. z)
        # if they are both moving along the same physical dimension
        self._coupled_axes = scope["stage"].get("coupled_axes", None)

        # print(f"Coupled axes: {self._coupled_axes} {type(self._coupled_axes)}")

        # safety
        assert (self._coupled_axes is None) or isinstance(self._coupled_axes, DictProxy)

        # If we have additional axes, create self.d{axis} for each
        # additional axis, to ensure we keep track of the step size
        if self._coupled_axes is not None:
            for leader, follower in self._coupled_axes.items():
                # print(leader, follower)
                assert leader.lower() in "xyzct"  # safety
                if getattr(self, f"d{follower.lower()}", None) is None:
                    # print(state.get(f"{follower.lower()}_step_size", 1))
                    setattr(
                        self,
                        f"d{follower.lower()}",
                        state.get(f"{follower.lower()}_step_size", 1),
                    )

    def set_stack_order_from_configuration_experiment(self) -> None:
        """Set stack order from configuration experiment"""
        state = self.configuration["experiment"]["MicroscopeState"]
        self._per_stack = (
            (
                state["stack_cycling_mode"] == "per_stack"
                and state["image_mode"] == "z-stack"
            )
            or (
                state["conpro_cycling_mode"] == "per_stack"
                and state["image_mode"] == "confocal-projection"
            )
            or (
                state["stack_cycling_mode"] == "per_stack"
                and state["image_mode"] == "ConstantVelocityAcquisition"
            )
        )

    @property
    def voxel_size(self) -> tuple:
        """Return voxel size

        Returns
        -------
        tuple
            Voxel size
        """
        return (self.dx, self.dy, self.dz)

    @property
    def shape(self) -> tuple:
        """Return shape as XYCZT.

        Returns
        -------
        tuple
            Shape as XYCZT
        """
        return (self.shape_x, self.shape_y, self.shape_c, self.shape_z, self.shape_t)


class XMLMetadata(Metadata):
    """XML Metadata

    This is a base class for dealing with metadata that is stored as an XML, e.g. in
    OME-TIFF or BigDataViewer. There are multiple methods for storing XML. In OME-TIFF,
    the XML is stored in the header of the first OME-TIFF file in a directory. In
    BigDataViewer, it is stored in a separate XML. Both have proprietary XML formats. To
    address this, we store their metadata in a nested dictionary, defined by
    {file_type}_xml_dict() (e.g. ome_tiff_xml_dict()) (see to_xml()). We then parse this
    nested dictionary into an XML file. Similarly, we can parse from an XML file to a
    nested dictionary and map these values back to our internal representation of
    metadata values (TODO: Not implemented. Will use xml_tools.parse_xml() for the
    first bit.)
    """

    def write_xml(
        self, file_name: str, file_type: str, root: Optional[str] = None, **kw
    ) -> None:
        """Write to XML file. Assumes we do not include the XML header in our nested
        metadata dictionary.

        Parameters
        ----------
        file_name : str
            File name
        file_type : str
            File type
        root : Optional[str], optional
            Root, by default None
        """
        xml = '<?xml version="1.0" encoding="UTF-8"?>\n'  # XML file header
        xml += (
            f"<!-- Created by Navigate, "
            f"v{__version__}, "
            f"Commit {__commit__}, Dean Lab at UTSW -->\n"
        )
        # TODO: should os.path.basename be the default? Added this for BigDataViewer's
        # relative path.
        xml += self.to_xml(file_type, root, file_name=os.path.basename(file_name), **kw)
        file_name = os.path.splitext(file_name)[0] + ".xml"
        with open(file_name, "w") as fp:
            fp.write(xml)

    def to_xml(self, file_type: str, root: Optional[str] = None, **kw) -> str:
        """
        Convert stored metadata to XML

        Parameters
        ----------
        file_type : str
            File type
        root : Optional[str], optional
            Root, by default None
        **kw
            Keyword arguments

        Returns
        -------
        str
            XML string
        """
        xml = ""
        try:
            d = getattr(
                self, f"{file_type.lower().replace(' ','_').replace('-','_')}_xml_dict"
            )(**kw)
            xml = xml_tools.dict_to_xml(d, root)
        except AttributeError:
            logging.debug(
                f"Metadata Writer - I do not know how to export {file_type} "
                f"metadata to XML."
            )
        return xml



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import logging

from navigate.model.devices.lasers.laser_base import LaserBase

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class SyntheticLaser(LaserBase):
    """SyntheticLaser Class"""

    def __init__(self, microscope_name, device_connection, configuration, laser_id):
        """Initialize the SyntheticLaser class.

        Parameters
        ----------
        microscope_name : str
            The microscope name.
        device_connection : object
            The device connection object.
        configuration : dict
            The device configuration.
        laser_id : str
            The laser ID.
        """
        super().__init__(microscope_name, device_connection, configuration, laser_id)

    def close(self):
        """Close the port before exit."""
        pass

    def initialize_laser(self):
        """
        Initialize lasers.
        """
        pass



import time
import numpy as np

from pyvcam import pvc
from pyvcam.camera import Camera
from pyvcam import constants

def main():
    pvc.init_pvcam()
    cam = next(Camera.detect_camera())
    cam.open()

    NUM_FRAMES = 10
    cnt = 0

    cam.start_seq(exp_time=20, num_frames=NUM_FRAMES)
    while cnt < NUM_FRAMES:
        frame, fps, frame_count = cam.poll_frame()

        low = np.amin(frame['pixel_data'])
        high = np.amax(frame['pixel_data'])
        average = np.average(frame['pixel_data'])

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Rate: {:.1f}\tFrame Count: {:.0f}\n'.format(low, high, average, fps, frame_count))
        cnt += 1

        time.sleep(0.05)

    cam.finish()

    # Test basic sequence methods
    frames = cam.get_sequence(NUM_FRAMES)
    for frame in frames:
        low = np.amin(frame)
        high = np.amax(frame)
        average = np.average(frame)

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Count: {:.0f}\n'.format(low, high, average, cnt))
        cnt += 1

    time_list = [i*10 for i in range(1, NUM_FRAMES+1)]
    frames = cam.get_vtm_sequence(time_list, constants.EXP_RES_ONE_MILLISEC, NUM_FRAMES)
    for frame in frames:
        low = np.amin(frame)
        high = np.amax(frame)
        average = np.average(frame)

        print('Min:{}\tMax:{}\tAverage:{:.0f}\tFrame Count: {:.0f}\n'.format(low, high, average, cnt))
        cnt += 1

    cam.close()
    pvc.uninit_pvcam()

    print('Total frames: {}\n'.format(cnt))


if __name__ == "__main__":
    main()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import xml.etree.ElementTree as ET


def dict_to_xml(d, tag=None, level=0):
    """Parse a Python dictionary to XML.

    Parameters
    ----------
    d: dict
        Dictionary to parse to XML.
    tag : str
        Root key of dictionary

    Returns
    -------
    xml : str
        String of XML tags produced from dictionary.
    """

    if tag is None:
        tag = list(d.keys())[0]

    xml = "  " * level + f"<{tag}"
    if isinstance(d, dict):
        next_xml = ""
        text = ""
        next_level = level + 1
        for k, v in d.items():
            if isinstance(v, dict):
                # Not a leaf node
                next_xml += dict_to_xml(v, k, next_level)
            elif isinstance(v, list):
                for el in v:
                    next_xml += dict_to_xml(el, k, next_level)
            else:
                if k == "text":
                    text = str(v)
                else:
                    xml += f' {k}="{v}"'
        if text != "" or next_xml != "":
            xml += ">"
            if text != "":
                xml += text
            if next_xml != "":
                xml += f"\n{next_xml}"
            if text != "":
                xml += f"</{tag}>\n"
            else:
                xml += "  " * level
                xml += f"</{tag}>\n"
        else:
            xml += "/>\n"

    return xml


def parse_xml(root: ET.Element) -> dict:
    """Parse an XML ElementTree.

    TODO: Does not account for namespacing. See OME-XML.

    Parameters
    ----------
    root : xml.etree.ElementTree.Element
        root Element of XML ElementTree

    Returns
    -------
    d : dict
        Dictionary representation of the XML file.
    """
    d = {}
    for k, v in root.attrib.items():
        d[k] = v
    try:
        text = root.text.strip()
        if text != "":
            d["text"] = text
    except AttributeError:
        # root.text is None
        pass
    prev_tag = ""
    for child in root:
        tag = child.tag
        if tag == prev_tag:
            if not isinstance(d[tag], list):
                # create the list
                tmp = d[tag]
                d[tag] = []
                d[tag].append(tmp)
            d[tag].append(parse_xml(child))
        else:
            d[tag] = parse_xml(child)
        prev_tag = tag
    return d
